# -*- coding: utf-8 -*-
# pylint: disable=fixme, invalid-name

# Disable while we have Python 2.x compatability
# pylint: disable=useless-object-inheritance

"""Classes representing Sonos UPnP services.

>>> import soco
>>> device = soco.SoCo('192.168.1.102')
>>> print(RenderingControl(device).GetMute([('InstanceID', 0),
...     ('Channel', 'Master')]))
{'CurrentMute': '0'}
>>> r = ContentDirectory(device).Browse([
...    ('ObjectID', 'Q:0'),
...    ('BrowseFlag', 'BrowseDirectChildren'),
...    ('Filter', '*'),
...    ('StartingIndex', '0'),
...    ('RequestedCount', '100'),
...    ('SortCriteria', '')
...    ])
>>> print(r['Result'])
<?xml version="1.0" ?><DIDL-Lite xmlns="urn:schemas-upnp-org:metadata ...
>>> for action, in_args, out_args in AlarmClock(device).iter_actions():
...    print(action, in_args, out_args)
...
SetFormat [Argument(name='DesiredTimeFormat', vartype='string'), Argument(
name='DesiredDateFormat', vartype='string')] []
GetFormat [] [Argument(name='CurrentTimeFormat', vartype='string'),
Argument(name='CurrentDateFormat', vartype='string')] ...
"""

# UPnP Spec at http://upnp.org/specs/arch/UPnP-arch-DeviceArchitecture-v1.0.pdf

from __future__ import (
    absolute_import, unicode_literals
)

import logging
from collections import namedtuple
from xml.sax.saxutils import escape

import requests

from .cache import Cache
from .events import Subscription
from .exceptions import (
    SoCoException, SoCoUPnPException, UnknownSoCoException
)
from .utils import prettify
from .xml import XML, illegal_xml_re

# UNICODE NOTE
# UPnP requires all XML to be transmitted/received with utf-8 encoding. All
# strings used in this module are unicode. The Requests library should take
# care of all of the necessary encoding (on sending) and decoding (on
# receiving) for us, provided that we specify the correct encoding headers
# (which, hopefully, we do).
# But since ElementTree seems to prefer being fed bytes to unicode, at least
# for Python 2.x, we have to encode strings specifically before using it. see
# http://bugs.python.org/issue11033 TODO: Keep an eye on this when it comes to
# Python 3 compatibility


log = logging.getLogger(__name__)  # pylint: disable=C0103
# logging.basicConfig()
# log.setLevel(logging.INFO)


class Action(namedtuple('ActionBase', 'name, in_args, out_args')):
    """A UPnP Action and its arguments."""
    def __str__(self):
        args = ', '.join(str(arg) for arg in self.in_args)
        returns = ', '.join(str(arg) for arg in self.out_args)
        return '{0}({1}) -> {{{2}}}'.format(self.name, args, returns)


class Argument(namedtuple('ArgumentBase', 'name, vartype')):
    """A UPnP Argument and its type."""
    def __str__(self):
        argument = self.name
        if self.vartype.default:
            argument = "{0}={1}".format(self.name, self.vartype.default)
        return '{0}: {1}'.format(argument, str(self.vartype))


class Vartype(namedtuple('VartypeBase', 'datatype, default, list, range')):
    """An argument type with default value and range."""
    def __str__(self):
        if self.list:
            return '[{0}]'.format(', '.join(self.list))
        if self.range:
            return '[{0}..{1}]'.format(self.range[0], self.range[1])
        return self.datatype


# pylint: disable=too-many-instance-attributes
class Service(object):

    """A class representing a UPnP service.

    This is the base class for all Sonos Service classes. This class has a
    dynamic method dispatcher. Calls to methods which are not explicitly
    defined here are dispatched automatically to the service action with the
    same name.
    """
    # pylint: disable=bad-continuation
    soap_body_template = (
        '<?xml version="1.0"?>'
        '<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"'
        ' s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">'
            '<s:Body>'
                '<u:{action} xmlns:u="urn:schemas-upnp-org:service:'
                    '{service_type}:{version}">'
                    '{arguments}'
                '</u:{action}>'
            '</s:Body>'
        '</s:Envelope>')  # noqa PEP8

    def __init__(self, soco):
        """
        Args:
            soco (SoCo): A `SoCo` instance to which the UPnP Actions will be
                sent
        """

        #: `SoCo`: The `SoCo` instance to which UPnP Actions are sent
        self.soco = soco
        # Some defaults. Some or all these will need to be overridden
        # specifically in a sub-class. There is other information we could
        # record, but this will do for the moment. Info about a Sonos device is
        # available at <IP_address>/xml/device_description.xml in the
        # <service> tags

        #: str: The UPnP service type.
        self.service_type = self.__class__.__name__
        #: str: The UPnP service version.
        self.version = 1
        self.service_id = self.service_type
        #: str: The base URL for sending UPnP Actions.
        self.base_url = 'http://{}:1400'.format(self.soco.ip_address)
        #: str: The UPnP Control URL.
        self.control_url = '/{}/Control'.format(self.service_type)
        #: str: The service control protocol description URL.
        self.scpd_url = '/xml/{}{}.xml'.format(
            self.service_type, self.version)
        #: str: The service eventing subscription URL.
        self.event_subscription_url = '/{}/Event'.format(self.service_type)
        #: A cache for storing the result of network calls. By default, this is
        #: a `TimedCache` with a default timeout=0.
        self.cache = Cache(default_timeout=0)

        # Caching variables for actions and event_vars, will be filled when
        # they are requested for the first time
        self._actions = None
        self._event_vars = None

        # From table 3.3 in
        # http://upnp.org/specs/arch/UPnP-arch-DeviceArchitecture-v1.1.pdf
        # This list may not be complete, but should be good enough to be going
        # on with.  Error codes between 700-799 are defined for particular
        # services, and may be overriden in subclasses. Error codes >800
        # are generally SONOS specific. NB It may well be that SONOS does not
        # use some of these error codes.

        # pylint: disable=invalid-name
        self.UPNP_ERRORS = {
            400: 'Bad Request',
            401: 'Invalid Action',
            402: 'Invalid Args',
            404: 'Invalid Var',
            412: 'Precondition Failed',
            501: 'Action Failed',
            600: 'Argument Value Invalid',
            601: 'Argument Value Out of Range',
            602: 'Optional Action Not Implemented',
            603: 'Out Of Memory',
            604: 'Human Intervention Required',
            605: 'String Argument Too Long',
            606: 'Action Not Authorized',
            607: 'Signature Failure',
            608: 'Signature Missing',
            609: 'Not Encrypted',
            610: 'Invalid Sequence',
            611: 'Invalid Control URL',
            612: 'No Such Session',
        }
        self.DEFAULT_ARGS = {}

    def __getattr__(self, action):
        """Called when a method on the instance cannot be found.

        Causes an action to be sent to UPnP server. See also
        `object.__getattr__`.

        Args:
            action (str): The name of the unknown method.
        Returns:
            callable: The callable to be invoked. .
        """

        # Define a function to be invoked as the method, which calls
        # send_command.
        def _dispatcher(self, *args, **kwargs):
            """Dispatch to send_command."""
            return self.send_command(action, *args, **kwargs)

        # rename the function so it appears to be the called method. We
        # probably don't need this, but it doesn't harm
        _dispatcher.__name__ = action

        # _dispatcher is now an unbound menthod, but we need a bound method.
        # This turns an unbound method into a bound method (i.e. one that
        # takes self - an instance of the class - as the first parameter)
        # pylint: disable=no-member
        method = _dispatcher.__get__(self, self.__class__)
        # Now we have a bound method, we cache it on this instance, so that
        # next time we don't have to go through this again
        setattr(self, action, method)
        log.debug("Dispatching method %s", action)

        # return our new bound method, which will be called by Python
        return method

    @staticmethod
    def wrap_arguments(args=None):
        """Wrap a list of tuples in xml ready to pass into a SOAP request.

        Args:
            args (list):  a list of (name, value) tuples specifying the
                name of each argument and its value, eg
                ``[('InstanceID', 0), ('Speed', 1)]``. The value
                can be a string or something with a string representation. The
                arguments are escaped and wrapped in <name> and <value> tags.

        Example:

            >>> from soco import SoCo
            >>> device = SoCo('192.168.1.101')
            >>> s = Service(device)
            >>> print(s.wrap_arguments([('InstanceID', 0), ('Speed', 1)]))
            <InstanceID>0</InstanceID><Speed>1</Speed>'
            """
        if args is None:
            args = []

        tags = []
        for name, value in args:
            tag = "<{name}>{value}</{name}>".format(
                name=name, value=escape("%s" % value, {'"': "&quot;"}))
            # % converts to unicode because we are using unicode literals.
            # Avoids use of 'unicode' function which does not exist in python 3
            tags.append(tag)

        xml = "".join(tags)
        return xml

    @staticmethod
    def unwrap_arguments(xml_response):
        """Extract arguments and their values from a SOAP response.

        Args:
            xml_response (str):  SOAP/xml response text (unicode,
                not utf-8).
        Returns:
             dict: a dict of ``{argument_name: value}`` items.
        """

        # A UPnP SOAP response (including headers) looks like this:

        # HTTP/1.1 200 OK
        # CONTENT-LENGTH: bytes in body
        # CONTENT-TYPE: text/xml; charset="utf-8" DATE: when response was
        # generated
        # EXT:
        # SERVER: OS/version UPnP/1.0 product/version
        #
        # <?xml version="1.0"?>
        # <s:Envelope
        #   xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"
        #   s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
        #   <s:Body>
        #       <u:actionNameResponse
        #           xmlns:u="urn:schemas-upnp-org:service:serviceType:v">
        #           <argumentName>out arg value</argumentName>
        #               ... other out args and their values go here, if any
        #       </u:actionNameResponse>
        #   </s:Body>
        # </s:Envelope>

        # Get all tags in order. Elementree (in python 2.x) seems to prefer to
        # be fed bytes, rather than unicode
        xml_response = xml_response.encode('utf-8')
        try:
            tree = XML.fromstring(xml_response)
        except XML.ParseError:
            # Try to filter illegal xml chars (as unicode), in case that is
            # the reason for the parse error
            filtered = illegal_xml_re.sub('', xml_response.decode('utf-8'))\
                                     .encode('utf-8')
            tree = XML.fromstring(filtered)

        # Get the first child of the <Body> tag which will be
        # <{actionNameResponse}> (depends on what actionName is). Turn the
        # children of this into a {tagname, content} dict. XML unescaping
        # is carried out for us by elementree.
        action_response = tree.find(
            "{http://schemas.xmlsoap.org/soap/envelope/}Body")[0]
        return dict((i.tag, i.text or "") for i in action_response)

    def compose_args(self, action_name, in_argdict):
        """Compose the argument list from an argument dictionary, with
        respect for default values.

        Args:
            action_name (str): The name of the action to be performed.
            in_argdict (dict): Arguments as a dict, eg
                ``{'InstanceID': 0, 'Speed': 1}. The values
                can be a string or something with a string representation.

        Returns:
            list: a list of ``(name, value)`` tuples.

        Raises:
            `AttributeError`: If this service does not support the action.
            `ValueError`: If the argument lists do not match the action
                signature.
        """

        for action in self.actions:
            if action.name == action_name:
                # The found 'action' will be visible from outside the loop
                break
        else:
            raise AttributeError('Unknown Action: {0}'.format(action_name))

        # Check for given argument names which do not occur in the expected
        # argument list
        # pylint: disable=undefined-loop-variable
        unexpected = set(in_argdict) - \
            set(argument.name for argument in action.in_args)
        if unexpected:
            raise ValueError(
                "Unexpected argument '{0}'. Method signature: {1}"
                .format(next(iter(unexpected)), str(action))
            )

        # List the (name, value) tuples for each argument in the argument list
        composed = []
        for argument in action.in_args:
            name = argument.name
            if name in in_argdict:
                composed.append((name, in_argdict[name]))
                continue
            if name in self.DEFAULT_ARGS:
                composed.append((name, self.DEFAULT_ARGS[name]))
                continue
            if argument.vartype.default is not None:
                composed.append((name, argument.vartype.default))
            raise ValueError(
                "Missing argument '{0}'. Method signature: {1}"
                .format(argument.name, str(action))
            )
        return composed

    def build_command(self, action, args=None):
        """Build a SOAP request.

        Args:
            action (str): the name of an action (a string as specified in the
                service description XML file) to be sent.
            args (list, optional): Relevant arguments as a list of (name,
                value) tuples.

        Returns:
            tuple: a tuple containing the POST headers (as a dict) and a
                string containing the relevant SOAP body. Does not set
                content-length, or host headers, which are completed upon
                sending.
        """

        # A complete request should look something like this:

        # POST path of control URL HTTP/1.1
        # HOST: host of control URL:port of control URL
        # CONTENT-LENGTH: bytes in body
        # CONTENT-TYPE: text/xml; charset="utf-8"
        # SOAPACTION: "urn:schemas-upnp-org:service:serviceType:v#actionName"
        #
        # <?xml version="1.0"?>
        # <s:Envelope
        #   xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"
        #   s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
        #   <s:Body>
        #       <u:actionName
        #           xmlns:u="urn:schemas-upnp-org:service:serviceType:v">
        #           <argumentName>in arg value</argumentName>
        #           ... other in args and their values go here, if any
        #       </u:actionName>
        #   </s:Body>
        # </s:Envelope>

        arguments = self.wrap_arguments(args)
        body = self.soap_body_template.format(
            arguments=arguments, action=action, service_type=self.service_type,
            version=self.version)
        soap_action_template = \
            "urn:schemas-upnp-org:service:{service_type}:{version}#{action}"
        soap_action = soap_action_template.format(
            service_type=self.service_type, version=self.version,
            action=action)
        headers = {'Content-Type': 'text/xml; charset="utf-8"',
                   'SOAPACTION': soap_action}
        # Note that although we set the charset to utf-8 here, in fact the
        # body is still unicode. It will only be converted to bytes when it
        # is set over the network
        return (headers, body)

    def send_command(self, action, args=None, cache=None, cache_timeout=None,
                     **kwargs):
        """Send a command to a Sonos device.

        Args:
            action (str): the name of an action (a string as specified in the
                service description XML file) to be sent.
            args (list, optional): Relevant arguments as a list of (name,
                value) tuples, as an alternative to ``kwargs``.
            cache (Cache): A cache is operated so that the result will be
                stored for up to ``cache_timeout`` seconds, and a subsequent
                call with the same arguments within that period will be
                returned from the cache, saving a further network call. The
                cache may be invalidated or even primed from another thread
                (for example if a UPnP event is received to indicate that
                the state of the Sonos device has changed). If
                ``cache_timeout`` is missing or `None`, the cache will use a
                default value (which may be 0 - see `cache`). By default,
                the cache identified by the service's `cache` attribute will
                be used, but a different cache object may be specified in
                the `cache` parameter.
            kwargs: Relevant arguments for the command.

        Returns:
             dict: a dict of ``{argument_name, value}`` items.

        Raises:
            `AttributeError`: If this service does not support the action.
            `ValueError`: If the argument lists do not match the action
                signature.
            `SoCoUPnPException`: if a SOAP error occurs.
            `UnknownSoCoException`: if an unknonwn UPnP error occurs.
            `requests.exceptions.HTTPError`: if an http error occurs.

        """
        if args is None:
            args = self.compose_args(action, kwargs)
        if cache is None:
            cache = self.cache
        result = cache.get(action, args)
        if result is not None:
            log.debug("Cache hit")
            return result
        # Cache miss, so go ahead and make a network call
        headers, body = self.build_command(action, args)
        log.debug("Sending %s %s to %s", action, args, self.soco.ip_address)
        log.debug("Sending %s, %s", headers, prettify(body))
        # Convert the body to bytes, and send it.
        try:
            response = requests.post(
                self.base_url + self.control_url,
                headers=headers,
                data=body.encode('utf-8'),
                timeout=20,
            )
        except requests.exceptions.RequestException as ex:
            raise SoCoException('Connection error: ' + str(ex))

        log.debug("Received %s, %s", response.headers, response.text)
        status = response.status_code
        log.debug(
            "Received status %s from %s", status, self.soco.ip_address)
        if status == 200:
            # The response is good. Get the output params, and return them.
            # NB an empty dict is a valid result. It just means that no
            # params are returned. By using response.text, we rely upon
            # the requests library to convert to unicode for us.
            result = self.unwrap_arguments(response.text) or True
            # Store in the cache. There is no need to do this if there was an
            # error, since we would want to try a network call again.
            cache.put(result, action, args, timeout=cache_timeout)
            return result
        elif status == 500:
            # Internal server error. UPnP requires this to be returned if the
            # device does not like the action for some reason. The returned
            # content will be a SOAP Fault. Parse it and raise an error.
            self.handle_upnp_error(response.text)
        else:
            # Something else has gone wrong. Probably a network error. Let
            # Requests handle it
            response.raise_for_status()
        return None

    def handle_upnp_error(self, xml_error):
        """Disect a UPnP error, and raise an appropriate exception.

        Args:
            xml_error (str):  a unicode string containing the body of the
                UPnP/SOAP Fault response. Raises an exception containing the
                error code.
        """

        # An error code looks something like this:

        # HTTP/1.1 500 Internal Server Error
        # CONTENT-LENGTH: bytes in body
        # CONTENT-TYPE: text/xml; charset="utf-8"
        # DATE: when response was generated
        # EXT:
        # SERVER: OS/version UPnP/1.0 product/version

        # <?xml version="1.0"?>
        # <s:Envelope
        #   xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"
        #   s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
        #   <s:Body>
        #       <s:Fault>
        #           <faultcode>s:Client</faultcode>
        #           <faultstring>UPnPError</faultstring>
        #           <detail>
        #               <UPnPError xmlns="urn:schemas-upnp-org:control-1-0">
        #                   <errorCode>error code</errorCode>
        #                   <errorDescription>error string</errorDescription>
        #               </UPnPError>
        #           </detail>
        #       </s:Fault>
        #   </s:Body>
        # </s:Envelope>
        #
        # All that matters for our purposes is the errorCode.
        # errorDescription is not required, and Sonos does not seem to use it.

        # NB need to encode unicode strings before passing to ElementTree
        xml_error = xml_error.encode('utf-8')
        error = XML.fromstring(xml_error)
        log.debug("Error %s", xml_error)
        error_code = error.findtext(
            './/{urn:schemas-upnp-org:control-1-0}errorCode')
        if error_code is not None:
            description = self.UPNP_ERRORS.get(int(error_code), '')
            raise SoCoUPnPException(
                message='UPnP Error {} received: {} from {}'.format(
                    error_code, description, self.soco.ip_address),
                error_code=error_code,
                error_description=description,
                error_xml=xml_error
            )

        # Unknown error, so just return the entire response
        log.error("Unknown error received from %s", self.soco.ip_address)
        raise UnknownSoCoException(xml_error)

    def subscribe(
            self, requested_timeout=None, auto_renew=False, event_queue=None):
        """Subscribe to the service's events.

        Args:
            requested_timeout (int, optional): If requested_timeout is
                provided, a subscription valid for that
                number of seconds will be requested, but not guaranteed. Check
                `Subscription.timeout` on return to find out what period of
                validity is actually allocated.
            auto_renew (bool): If auto_renew is `True`, the subscription will
                automatically be renewed just before it expires, if possible.
                Default is `False`.

            event_queue (:class:`~queue.Queue`): a thread-safe queue object on
                which received events will be put. If not specified,
                a (:class:`~queue.Queue`) will be created and used.

        Returns:
            `Subscription`: an insance of `Subscription`, representing
                the new subscription.

        To unsubscribe, call the `unsubscribe` method on the returned object.
        """
        subscription = Subscription(
            self, event_queue)
        subscription.subscribe(
            requested_timeout=requested_timeout, auto_renew=auto_renew)
        return subscription

    def _update_cache_on_event(self, event):
        """Update the cache when an event is received.

        This will be called before an event is put onto the event queue. Events
        will often indicate that the Sonos device's state has changed, so this
        opportunity is made available for the service to update its cache. The
        event will be put onto the event queue once this method returns.

        `event` is an Event namedtuple: ('sid', 'seq', 'service', 'variables')

        ..  warning:: This method will not be called from the main thread but
            by one or more threads, which handle the events as they come in.
            You *must not* access any class, instance or global variables
            without appropriate locks. Treat all parameters passed to this
            method as read only.
        """

    @property
    def actions(self):
        """The service's actions with their arguments.

        Returns:
            list(`Action`): A list of Action namedtuples, consisting of
            action_name (str), in_args (list of Argument namedtuples,
            consisting of name and argtype), and out_args (ditto).

        The return value looks like this::
            [
                Action(
                    name='GetMute',
                    in_args=[
                        Argument(name='InstanceID', ...),
                        Argument(
                            name='Channel',
                            vartype='string',
                            list=['Master', 'LF', 'RF', 'SpeakerOnly'],
                            range=None
                        )
                    ],
                    out_args=[
                        Argument(name='CurrentMute, ...)
                    ]
                )
                Action(...)
            ]

        Its string representation will look like this::
            GetMute(InstanceID: ui4, Channel: [Master, LF, RF, SpeakerOnly]) \
            -> {CurrentMute: boolean}
        """
        if self._actions is None:
            self._actions = list(self.iter_actions())
        return self._actions

    def iter_actions(self):
        """Yield the service's actions with their arguments.

        Yields:
            `Action`: the next action.

        Each action is an Action namedtuple, consisting of action_name
        (a string), in_args (a list of Argument namedtuples consisting of name
        and argtype), and out_args (ditto), eg::

            Action(
                name='SetFormat',
                in_args=[
                    Argument(name='DesiredTimeFormat', vartype=<Vartype>),
                    Argument(name='DesiredDateFormat', vartype=<Vartype>)],
                out_args=[]
            )
        """

        # pylint: disable=too-many-locals
        # pylint: disable=invalid-name
        ns = '{urn:schemas-upnp-org:service-1-0}'
        # get the scpd body as bytes, and feed directly to elementtree
        # which likes to receive bytes
        scpd_body = requests.get(
            self.base_url + self.scpd_url, timeout=10).content
        tree = XML.fromstring(scpd_body)
        # parse the state variables to get the relevant variable types
        vartypes = {}
        srvStateTables = tree.findall('{}serviceStateTable'.format(ns))
        for srvStateTable in srvStateTables:
            statevars = srvStateTable.findall('{}stateVariable'.format(ns))
            for state in statevars:
                name = state.findtext('{}name'.format(ns))
                datatype = state.findtext('{}dataType'.format(ns))
                default = state.findtext('{}defaultValue'.format(ns))
                value_list_elt = state.find('{}allowedValueList'.format(ns))
                if value_list_elt is None:
                    value_list_elt = ()
                value_list = [item.text for item in value_list_elt] or None
                value_range_elt = state.find('{}allowedValueRange'.format(ns))
                if value_range_elt is None:
                    value_range_elt = ()
                value_range = [item.text for item in value_range_elt] or None
                vartypes[name] = Vartype(datatype, default, value_list,
                                         value_range)
        # find all the actions
        actionLists = tree.findall('{}actionList'.format(ns))
        for actionList in actionLists:
            actions = actionList.findall('{}action'.format(ns))
            for i in actions:
                action_name = i.findtext('{}name'.format(ns))
                argLists = i.findall('{}argumentList'.format(ns))
                for argList in argLists:
                    args_iter = argList.findall('{}argument'.format(ns))
                    in_args = []
                    out_args = []
                    for arg in args_iter:
                        arg_name = arg.findtext('{}name'.format(ns))
                        direction = arg.findtext('{}direction'.format(ns))
                        related_variable = arg.findtext(
                            '{}relatedStateVariable'.format(ns))
                        vartype = vartypes[related_variable]
                        if direction == "in":
                            in_args.append(Argument(arg_name, vartype))
                        else:
                            out_args.append(Argument(arg_name, vartype))
                    yield Action(action_name, in_args, out_args)

    @property
    def event_vars(self):
        """The service's eventable variables.

        Returns:
            list(tuple): A list of (variable name, data type) tuples.
        """
        if self._event_vars is None:
            self._event_vars = list(self.iter_event_vars())
        return self._event_vars

    def iter_event_vars(self):
        """Yield the services eventable variables.

        Yields:
            `tuple`: a tuple of (variable name, data type).
        """

        # pylint: disable=invalid-name
        ns = '{urn:schemas-upnp-org:service-1-0}'
        scpd_body = requests.get(
            self.base_url + self.scpd_url, timeout=10).text
        tree = XML.fromstring(scpd_body.encode('utf-8'))
        # parse the state variables to get the relevant variable types
        statevars = tree.findall('{}stateVariable'.format(ns))
        for state in statevars:
            # We are only interested if 'sendEvents' is 'yes', i.e this
            # is an eventable variable
            if state.attrib['sendEvents'] == "yes":
                name = state.findtext('{}name'.format(ns))
                vartype = state.findtext('{}dataType'.format(ns))
                yield (name, vartype)


class AlarmClock(Service):

    """Sonos alarm service, for setting and getting time and alarms."""

    def __init__(self, soco):
        super(AlarmClock, self).__init__(soco)
        self.UPNP_ERRORS.update(
            {
                801: 'Already an alarm for this time',
            })


class MusicServices(Service):

    """Sonos music services service, for functions related to 3rd party music
    services."""


class DeviceProperties(Service):

    """Sonos device properties service, for functions relating to zones, LED
    state, stereo pairs etc."""


class SystemProperties(Service):

    """Sonos system properties service, for functions relating to
    authentication etc."""


class ZoneGroupTopology(Service):

    """Sonos zone group topology service, for functions relating to network
    topology, diagnostics and updates."""


class GroupManagement(Service):

    """Sonos group management service, for services relating to groups."""


class QPlay(Service):

    """Sonos Tencent QPlay service (a Chinese music service)"""


class ContentDirectory(Service):

    """UPnP standard Content Directory service, for functions relating to
    browsing, searching and listing available music."""

    def __init__(self, soco):
        super(ContentDirectory, self).__init__(soco)
        self.control_url = "/MediaServer/ContentDirectory/Control"
        self.event_subscription_url = "/MediaServer/ContentDirectory/Event"
        # For error codes, see table 2.7.16 in
        # http://upnp.org/specs/av/UPnP-av-ContentDirectory-v1-Service.pdf
        self.UPNP_ERRORS.update({
            701: 'No such object',
            702: 'Invalid CurrentTagValue',
            703: 'Invalid NewTagValue',
            704: 'Required tag',
            705: 'Read only tag',
            706: 'Parameter Mismatch',
            708: 'Unsupported or invalid search criteria',
            709: 'Unsupported or invalid sort criteria',
            710: 'No such container',
            711: 'Restricted object',
            712: 'Bad metadata',
            713: 'Restricted parent object',
            714: 'No such source resource',
            715: 'Resource access denied',
            716: 'Transfer busy',
            717: 'No such file transfer',
            718: 'No such destination resource',
            719: 'Destination resource access denied',
            720: 'Cannot process the request',
        })


class MS_ConnectionManager(Service):  # pylint: disable=invalid-name

    """UPnP standard connection manager service for the media server."""

    def __init__(self, soco):
        super(MS_ConnectionManager, self).__init__(soco)
        self.service_type = "ConnectionManager"
        self.control_url = "/MediaServer/ConnectionManager/Control"
        self.event_subscription_url = "/MediaServer/ConnectionManager/Event"


class RenderingControl(Service):

    """UPnP standard rendering control service, for functions relating to
    playback rendering, eg bass, treble, volume and EQ."""

    def __init__(self, soco):
        super(RenderingControl, self).__init__(soco)
        self.control_url = "/MediaRenderer/RenderingControl/Control"
        self.event_subscription_url = "/MediaRenderer/RenderingControl/Event"
        self.DEFAULT_ARGS.update({'InstanceID': 0})


class MR_ConnectionManager(Service):  # pylint: disable=invalid-name

    """UPnP standard connection manager service for the media renderer."""

    def __init__(self, soco):
        super(MR_ConnectionManager, self).__init__(soco)
        self.service_type = "ConnectionManager"
        self.control_url = "/MediaRenderer/ConnectionManager/Control"
        self.event_subscription_url = "/MediaRenderer/ConnectionManager/Event"


class AVTransport(Service):

    """UPnP standard AV Transport service, for functions relating to transport
    management, eg play, stop, seek, playlists etc."""

    def __init__(self, soco):
        super(AVTransport, self).__init__(soco)
        self.control_url = "/MediaRenderer/AVTransport/Control"
        self.event_subscription_url = "/MediaRenderer/AVTransport/Event"
        # For error codes, see
        # http://upnp.org/specs/av/UPnP-av-AVTransport-v1-Service.pdf
        self.UPNP_ERRORS.update({
            701: 'Transition not available',
            702: 'No contents',
            703: 'Read error',
            704: 'Format not supported for playback',
            705: 'Transport is locked',
            706: 'Write error',
            707: 'Media is protected or not writeable',
            708: 'Format not supported for recording',
            709: 'Media is full',
            710: 'Seek mode not supported',
            711: 'Illegal seek target',
            712: 'Play mode not supported',
            713: 'Record quality not supported',
            714: 'Illegal MIME-Type',
            715: 'Content "BUSY"',
            716: 'Resource Not found',
            717: 'Play speed not supported',
            718: 'Invalid InstanceID',
            737: 'No DNS Server',
            738: 'Bad Domain Name',
            739: 'Server Error',
        })
        self.DEFAULT_ARGS.update({'InstanceID': 0})


class Queue(Service):

    """Sonos queue service, for functions relating to queue management, saving
    queues etc."""

    def __init__(self, soco):
        super(Queue, self).__init__(soco)
        self.control_url = "/MediaRenderer/Queue/Control"
        self.event_subscription_url = "/MediaRenderer/Queue/Event"


class GroupRenderingControl(Service):

    """Sonos group rendering control service, for functions relating to group
    volume etc."""

    def __init__(self, soco):
        super(GroupRenderingControl, self).__init__(soco)
        self.control_url = "/MediaRenderer/GroupRenderingControl/Control"
        self.event_subscription_url = \
            "/MediaRenderer/GroupRenderingControl/Event"
        self.DEFAULT_ARGS.update({'InstanceID': 0})



# -*- coding: utf-8 -*-

"""The Spotify plugin has been DEPRECATED

The Spotify Plugin has been immediately deprecated (August 2016),
because the API it was based on (The Spotify Metadata API) has been
ended. Since this rendered the plug-in broken, there was no need to
forewarn of the deprecation.

Please consider moving to the new general music services code (in
soco.music_services.music_service), that makes it possible to
retrived information about the available media from all music
services. A short intro for how to use the new code is available
in the API documentation:

 * http://docs.python-soco.com/en/latest/api/\
soco.music_services.music_service.html

and for some information about how to add items from the music
services to the queue, see this gist:

 * https://gist.github.com/lawrenceakka/2d21dca590b4fa7e3af2"

This deprecation notification will be deleted for the second release
after 0.12.

"""

import sys
import os
# Only raise this import error if we are not building the docs
if not (os.environ.get('READTHEDOCS', None) == 'True' or
        'sphinx' in sys.modules):
    raise RuntimeError(__doc__)



# -*- coding: utf-8 -*-
# pylint: disable=star-args,no-member

"""Unit tests for the music service data structures."""

from __future__ import unicode_literals

from xml.sax.saxutils import escape

import pytest

from pysonos.exceptions import DIDLMetadataError
from pysonos.ms_data_structures import (
    MSAlbum, MSAlbumList, MSArtist, MSArtistTracklist,
    MSCollection, MSFavorites, MSPlaylist, MSTrack
)
from pysonos.xml import XML

##############################################################################
# Example XML and the content dict to compare with for MS items              #
##############################################################################
DIDL_TEMPLATE = """
<DIDL-Lite xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
 xmlns:dc="http://purl.org/dc/elements/1.1/"
 xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/"
 xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">
<item id="{extended_id}" parentID="{parent_id}" restricted="true">
<dc:title>{title}</dc:title>
<upnp:class>{item_class}</upnp:class>
<desc id="cdudn"
 nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">{description}</desc>
</item>
</DIDL-Lite>
"""
DIDL_TEMPLATE = DIDL_TEMPLATE.replace('\n', '')

MS_TRACK_SEARCH_XML = """
<ns0:mediaMetadata xmlns:ns0="http://www.sonos.com/Services/1.1">
<ns0:id>trackid_24125935</ns0:id>
<ns0:itemType>track</ns0:itemType>
<ns0:mimeType>audio/aac</ns0:mimeType>
<ns0:title>Pilgrim</ns0:title>
<ns0:trackMetadata>
<ns0:artistId>artistid_4816276</ns0:artistId>
<ns0:artist>MØ</ns0:artist>
<ns0:composerId>artistid_4816276</ns0:composerId>
<ns0:composer>MØ</ns0:composer>
<ns0:albumId>albumid_24125922</ns0:albumId>
<ns0:album>Nytårsfesten 2014</ns0:album>
<ns0:albumArtistId>artistid_4816276</ns0:albumArtistId>
<ns0:albumArtist>MØ</ns0:albumArtist>
<ns0:duration>231</ns0:duration>
<ns0:albumArtURI>http://varnish01.music.aspiro.com/sca/imscale?h=90&amp;
w=90&amp;img=/content/sg19/vd00/music/prod/sonyddex/1386109801/
A10301A00030834072_20131203071914171/resources/A10301A00030834072_T-10764
_Image.jpg</ns0:albumArtURI>
<ns0:canPlay>true</ns0:canPlay>
<ns0:canSkip>true</ns0:canSkip>
<ns0:canAddToFavorites>true</ns0:canAddToFavorites>
</ns0:trackMetadata>
</ns0:mediaMetadata>
"""
MS_TRACK_SEARCH_XML = MS_TRACK_SEARCH_XML.replace('\n', '')
MS_TRACK_SEARCH_DICT = {
    'can_add_to_favorites': True,
    'composer': 'M\xd8',
    'album_id': 'albumid_24125922',
    'can_skip': True,
    'uri': 'x-sonos-http:trackid_24125935.mp4?sid=20&flags=32',
    'item_id': 'trackid_24125935',
    'item_type': 'track',
    'extended_id': '00030020trackid_24125935',
    'duration': 231,
    'can_play': True,
    'composer_id': 'artistid_4816276',
    'description': 'SA_RINCON5127_4542255535',
    'album': 'Nyt\xe5rsfesten 2014',
    'title': 'Pilgrim',
    'artist': 'M\xd8',
    'album_artist_id': 'artistid_4816276',
    'album_art_uri': 'http://varnish01.music.aspiro.com/sca/imscale?'
                     'h=90&w=90&img=/content/sg19/vd00/music/prod/sonyddex'
                     '/1386109801/A10301A00030834072_20131203071914171/'
                     'resources/A10301A00030834072_T-10764_Image.jpg',
    'album_artist': 'M\xd8',
    'parent_id': '00020064tracksearch:pilgrim',
    'service_id': 20,
    'artist_id': 'artistid_4816276',
    'mime_type': 'audio/aac'
}
MS_ALBUM_SEARCH_XML = """
<ns0:mediaCollection readOnly="true"
 xmlns:ns0="http://www.sonos.com/Services/1.1">
<ns0:id>albumid_5738780</ns0:id>
<ns0:itemType>album</ns0:itemType>
<ns0:title>Greatest De Unge År</ns0:title>
<ns0:artist>tv·2</ns0:artist>
<ns0:canPlay>true</ns0:canPlay>
<ns0:canAddToFavorites>true</ns0:canAddToFavorites>
<ns0:albumArtURI>http://varnish01.music.aspiro.com/sca/imscale?h=90&amp;w=90
&amp;img=/content/music8/prod/sonybmg/content/00000000000002304180/000/000/
000/000/056/501/97/00000000000005650197-1000x1000_72dpi_RGB_100Q.jpg
</ns0:albumArtURI>
</ns0:mediaCollection>
"""
MS_ALBUM_SEARCH_XML = MS_ALBUM_SEARCH_XML.replace('\n', '')
MS_ALBUM_SEARCH_DICT = {
    'can_add_to_favorites': True,
    'description': 'SA_RINCON5127_4542255535',
    'artist': 'tv\xb72',
    'title': 'Greatest De Unge \xc5r',
    'album_art_uri': 'http://varnish01.music.aspiro.com/sca/imscale?h=90&w=90'
                     '&img=/content/music8/prod/sonybmg/content/'
                     '00000000000002304180/000/000/000/000/056/501/97/'
                     '00000000000005650197-1000x1000_72dpi_RGB_100Q.jpg',
    'uri': 'x-rincon-cpcontainer:0004002calbumid_5738780',
    'parent_id': '00020064albumsearch:de unge',
    'item_type': 'album',
    'extended_id': '0004002calbumid_5738780',
    'item_id': 'albumid_5738780',
    'service_id': 20,
    'can_play': True
}
MS_ARTIST_SEARCH_XML = """
<ns0:mediaCollection readOnly="true"
 xmlns:ns0="http://www.sonos.com/Services/1.1">
<ns0:id>artistid_4761386</ns0:id>
<ns0:itemType>artist</ns0:itemType>
<ns0:title>Fritjof Såheim</ns0:title>
<ns0:artist>Fritjof Såheim</ns0:artist>
<ns0:canAddToFavorites>true</ns0:canAddToFavorites>
<ns0:albumArtURI>http://varnish01.music.aspiro.com/im/im?h=42&amp;w=64&amp;
artistid=4761386</ns0:albumArtURI>
</ns0:mediaCollection>
"""
MS_ARTIST_SEARCH_XML = MS_ARTIST_SEARCH_XML.replace('\n', '')
MS_ARTIST_SEARCH_DICT = {
    'can_add_to_favorites': True,
    'description': 'SA_RINCON5127_4542255535',
    'artist': 'Fritjof S\xe5heim',
    'title': 'Fritjof S\xe5heim',
    'album_art_uri': 'http://varnish01.music.aspiro.com/im/im?h=42&w=64'
                     '&artistid=4761386',
    'parent_id': '00020064artistsearch:Fritjof',
    'item_type': 'artist',
    'extended_id': '10050024artistid_4761386',
    'item_id': 'artistid_4761386',
    'service_id': 20
}
MS_PLAYLIST_SEARCH_XML = """
<ns0:mediaCollection readOnly="true" xmlns:ns0="http://www.sonos.com/Services/1.1">
  <ns0:id>playlistid_133fe1ff-1f16-4300-b440-fb80573f19ce</ns0:id>
  <ns0:itemType>albumList</ns0:itemType>
  <ns0:title>Kunstnerliste: Dans &amp; Lær</ns0:title>
  <ns0:artist>39 sange</ns0:artist>
  <ns0:artistId>artistid_0</ns0:artistId>
  <ns0:canPlay>true</ns0:canPlay>
  <ns0:canEnumerate>true</ns0:canEnumerate>
  <ns0:canAddToFavorites>true</ns0:canAddToFavorites>
  <ns0:albumArtURI>http://varnish01.music.aspiro.com/im/im?h=160&amp;w=240&amp;rows=1&amp;cols=2&amp;artimg&amp;uuid=133fe1ff-1f16-4300-b440-fb80573f19ce</ns0:albumArtURI>
</ns0:mediaCollection>
"""
MS_PLAYLIST_SEARCH_XML = MS_PLAYLIST_SEARCH_XML.replace('\n', '')
MS_PLAYLIST_SEARCH_DICT = {
    'can_add_to_favorites': True,
    'description': 'SA_RINCON5127_4542255535',
    'artist': '39 sange',
    'title': 'Kunstnerliste: Dans & L\xe6r',
    'album_art_uri': 'http://varnish01.music.aspiro.com/im/im?h=160&w=240&'
                     'rows=1&cols=2&artimg&uuid=133fe1ff-1f16-4300-b440-'
                     'fb80573f19ce',
    'uri': 'x-rincon-cpcontainer:000d006cplaylistid_133fe1ff-1f16-4300-b440'
           '-fb80573f19ce',
    'parent_id': '00020064playlistsearch:Dans &',
    'item_type': 'albumList',
    'extended_id': '000d006cplaylistid_133fe1ff-1f16-4300-b440-fb80573f19ce',
    'item_id': 'playlistid_133fe1ff-1f16-4300-b440-fb80573f19ce',
    'service_id': 20,
    'artist_id': 'artistid_0',
    'can_play': True,
    'can_enumerate': True
}


class FakeMusicService(object):
    """A fake music service."""

    def __init__(self, username):
        self.description = 'SA_RINCON5127_{}'.format(username)
        self.service_id = 20

    @staticmethod
    def id_to_extended_id(item_id, item_class):
        """ID to extended ID method."""
        id_prefix = {
            MSTrack: '00030020',
            MSAlbum: '0004002c',
            MSArtist: '10050024',
            MSAlbumList: '000d006c',
            MSPlaylist: '0006006c',
            MSArtistTracklist: '100f006c',
            MSFavorites: None,  # This one is unknown
            MSCollection: None  # This one is unknown
        }
        out = id_prefix[item_class]
        if out:
            out += item_id
        return out

    @staticmethod
    def form_uri(item_content, item_class):
        """Form the URI."""
        mime_type_to_extension = {'audio/aac': 'mp4'}
        uris = {
            MSTrack: 'x-sonos-http:{item_id}.{extension}?sid={service_id}&'
                     'flags=32',
            MSAlbum: 'x-rincon-cpcontainer:{extended_id}',
            MSAlbumList: 'x-rincon-cpcontainer:{extended_id}',
            MSPlaylist: 'x-rincon-cpcontainer:{extended_id}',
            MSArtistTracklist: 'x-rincon-cpcontainer:{extended_id}'
        }
        extension = None
        if 'mime_type' in item_content:
            extension = mime_type_to_extension[item_content['mime_type']]
        out = uris.get(item_class)
        if out:
            # pylint: disable=star-args
            out = out.format(extension=extension, **item_content)
        return out


FAKE_MUSIC_SERVICE = FakeMusicService('4542255535')


def getter_attributes_test(name, from_xml, from_dict, result):
    """Test if the getters return the right value."""
    assert getattr(from_xml, name) == result
    assert getattr(from_dict, name) == result


def common_tests(class_, xml_, dict_, parent_id):
    """Common tests for the MS classes."""
    xml_content = XML.fromstring(xml_.encode('utf8'))

    # MusicServiceItem.from_xml and MusicServiceItem.to_dict
    item_from_xml = class_.from_xml(xml_content, FAKE_MUSIC_SERVICE, parent_id)
    assert item_from_xml.to_dict == dict_

    # MusicServiceItem.from_dict and MusicServiceItem.to_dict
    item_from_dict = class_.from_dict(dict_)
    assert item_from_dict.to_dict == dict_

    # MusicServiceItem.didl_metadata
    # NOTE! These tests is reliant on the attributes being put in the same
    # order by ElementTree and so it might fail if that changes
    if item_from_xml.can_play:
        dict_encoded = {}
        for key, value in dict_.items():
            try:
                is_str = isinstance(value, unicode)
            except NameError:
                is_str = isinstance(value, str)
            if is_str:
                dict_encoded[key] = escape(value).\
                    encode('ascii', 'xmlcharrefreplace').decode('ascii')

            else:
                dict_encoded[key] = value
        didl = DIDL_TEMPLATE.format(item_class=class_.item_class,
                                    **dict_encoded)
        assert XML.tostring(item_from_xml.didl_metadata).decode('ascii') == \
            didl
        assert XML.tostring(item_from_dict.didl_metadata).decode('ascii') == \
            didl
    else:
        with pytest.raises(DIDLMetadataError):
            # pylint: disable=pointless-statement
            item_from_xml.didl_metadata

    # Text attributes with mandatory content
    for name in ['item_id', 'extended_id', 'title', 'service_id']:
        getter_attributes_test(name, item_from_xml, item_from_dict,
                               dict_[name])
    # Text attributes with voluntary content
    for name in ['parent_id', 'album_art_uri']:
        getter_attributes_test(name, item_from_xml, item_from_dict,
                               dict_.get(name))
    # Boolean attribute
    getter_attributes_test('can_play', item_from_xml, item_from_dict,
                           bool(dict_.get('can_play')))
    return item_from_xml, item_from_dict


def test_ms_track_search():
    """Test the MSTrack item when instantiated from a search."""
    item_from_xml, item_from_dict = common_tests(
        MSTrack,
        MS_TRACK_SEARCH_XML,
        MS_TRACK_SEARCH_DICT,
        '00020064tracksearch:pilgrim'
    )
    getter_attributes_test('artist', item_from_xml, item_from_dict,
                           MS_TRACK_SEARCH_DICT.get('artist'))
    getter_attributes_test('uri', item_from_xml, item_from_dict,
                           MS_TRACK_SEARCH_DICT['uri'])


def test_ms_album_search():
    """Test the MSAlbum item when instantiated from a search."""
    item_from_xml, item_from_dict = common_tests(
        MSAlbum,
        MS_ALBUM_SEARCH_XML,
        MS_ALBUM_SEARCH_DICT,
        '00020064albumsearch:de unge'
    )
    getter_attributes_test('artist', item_from_xml, item_from_dict,
                           MS_ALBUM_SEARCH_DICT.get('artist'))
    getter_attributes_test('uri', item_from_xml, item_from_dict,
                           MS_ALBUM_SEARCH_DICT['uri'])


def test_ms_artist_search():
    """Test the MSAlbum item when instantiated from a search."""
    common_tests(
        MSArtist,
        MS_ARTIST_SEARCH_XML,
        MS_ARTIST_SEARCH_DICT,
        '00020064artistsearch:Fritjof'
    )


def test_ms_playlist_search():
    """Test the MSAlbum item when instantiated from a search."""
    item_from_xml, item_from_dict = common_tests(
        MSAlbumList,
        MS_PLAYLIST_SEARCH_XML,
        MS_PLAYLIST_SEARCH_DICT,
        '00020064playlistsearch:Dans &'
    )
    getter_attributes_test('uri', item_from_xml, item_from_dict,
                           MS_PLAYLIST_SEARCH_DICT['uri'])



# -*- coding: utf-8 -*-
"""Tests for the cache module."""

from __future__ import unicode_literals

from pysonos.cache import (
    Cache, NullCache, TimedCache
)


def test_instance_creation():
    assert isinstance(Cache(), TimedCache)
    from pysonos import config
    config.CACHE_ENABLED = False
    assert isinstance(Cache(), NullCache)
    config.CACHE_ENABLED = True


def test_cache_put_get():
    """Test putting items into, and getting them from, the cache."""
    from time import sleep
    cache = Cache()
    cache.put("item", 'some', kw='args', timeout=3)
    assert not cache.get('some', 'otherargs') == "item"
    assert cache.get('some', kw='args') == "item"
    sleep(2)
    assert cache.get('some', kw='args') == "item"
    sleep(2)
    assert not cache.get('some', kw='args') == "item"

    cache.put("item", 'some', 'args', and_a='keyword', timeout=3)
    assert cache.get('some', 'args', and_a='keyword') == "item"
    assert not cache.get(
        'some', 'otherargs', and_a='keyword') == "item"


def test_cache_clear_del():
    """Test removal of items and clearing the cache."""
    cache = Cache()
    cache.put("item", "some", kw="args", timeout=2)
    # Check it's there
    assert cache.get('some', kw='args') == "item"
    # delete it
    cache.delete('some', kw='args')
    assert not cache.get('some', kw='args') == "item"
    # put it back
    cache.put("item", "some", kw="args", timeout=3)
    cache.clear()
    assert not cache.get('some', kw='args') == "item"


def test_with_typical_args():
    cache = Cache()
    cache.put("result", 'SetAVTransportURI', [
        ('InstanceID', 1),
        ('CurrentURI', 'URI2'),
        ('CurrentURIMetaData', 'abcd'),
        ('Unicode', 'μИⅠℂ☺ΔЄ💋')
    ], timeout=3)
    assert cache.get('SetAVTransportURI', [
        ('InstanceID', 1),
        ('CurrentURI', 'URI2'),
        ('CurrentURIMetaData', 'abcd'),
        ('Unicode', 'μИⅠℂ☺ΔЄ💋')
    ]) == "result"


def test_cache_disable():
    cache = Cache()
    assert cache.enabled is True
    cache.enabled = False
    cache.put("item", 'args', timeout=3)
    assert cache.get('args') == None
    # Check it's there
    assert cache.get('some', kw='args') is None



# -*- coding: utf-8 -*-
"""Tests for the soap module."""

from __future__ import unicode_literals

from pysonos.soap import SoapMessage
from pysonos.xml import XML

try:
    # New in Python 3.3
    from unittest import mock
except ImportError:
    # Python 2.7
    import mock


DUMMY_VALID_RESPONSE = "".join([
    '<?xml version="1.0"?>',
    '<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"',
    ' s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">',
    '<s:Body>',
    '<u:GetLEDStateResponse ',
    'xmlns:u="urn:schemas-upnp-org:service:DeviceProperties:1">',
    '<CurrentLEDState>On</CurrentLEDState>',
    '<Unicode>data</Unicode>',
    '</u:GetLEDStateResponse>',
    '</s:Body>',
    '</s:Envelope>'])  # noqa PEP8


def test_init():
    """Tests for initialisation of classes."""
    s = SoapMessage('http://endpoint_url', 'a_method')
    assert s.endpoint == "http://endpoint_url"
    assert s.method == 'a_method'
    assert s.parameters == []
    assert s.soap_action is None
    assert s.http_headers is None
    assert s.soap_header is None
    assert s.namespace is None
    assert s.request_args == {}


def test_prepare_headers():
    """Check http_headers are correctly prepared."""
    s = SoapMessage('endpoint', 'method')
    h = s.prepare_headers({'test1': 'one', 'test2': 'two'}, None)
    assert h == {'Content-Type': 'text/xml; charset="utf-8"',
                 'test2': 'two', 'test1': 'one'}
    h = s.prepare_headers(http_headers={'test1': 'one', 'test2': 'two'},
                          soap_action='soapaction')
    assert h == {'Content-Type': 'text/xml; charset="utf-8"',
                 'test2': 'two', 'test1': 'one',
                 'SOAPACTION': '"soapaction"'}


def test_prepare_soap_header():
    """Check that the SOAP header is correctly wrapped."""
    s = SoapMessage('endpoint', 'method')
    h = s.prepare_soap_header('<a><b></b></a>')
    assert h == "<s:Header><a><b></b></a></s:Header>"
    s = SoapMessage('endpoint', 'method', http_headers={
        'test1': 'one', 'test2': 'two'})
    h = s.prepare_soap_header(None)
    assert h == ''


def test_prepare_soap_body():
    """Check that the SOAP body is correctly prepared."""
    # No params
    s = SoapMessage('endpoint', 'method')
    b = s.prepare_soap_body('a_method', [], None)
    assert b == "<a_method></a_method>"
    # One param
    b = s.prepare_soap_body('a_method', [('one', '1')], None)
    assert b == "<a_method><one>1</one></a_method>"
    # Two params
    b = s.prepare_soap_body('a_method', [
        ('one', '1'), ('two', '2')
    ], None)
    assert b == "<a_method><one>1</one><two>2</two></a_method>"

    # And with a namespace

    b = s.prepare_soap_body('a_method', [('one', '1'),
                                         ('two', '2')], "http://a_namespace")
    assert b == "<a_method " \
                'xmlns="http://a_namespace"><one>1</one><two>2</two' \
                '></a_method>'


def test_prepare():
    """Test preparation of whole SOAP message."""
    s = SoapMessage(
        endpoint='endpoint',
        method='getData',
        parameters=[('one', '1')],
        http_headers={'timeout': '3'},
        soap_action='ACTION',
        soap_header="<a_header>data</a_header>",
        namespace="http://namespace.com")
    headers, data = s.prepare()
    assert data == '<?xml version="1.0"?><s:Envelope ' \
                   'xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" ' \
                   's:encodingStyle="http://schemas.xmlsoap.org/soap' \
                   '/encoding/"><s:Header><a_header>data</a_header></s' \
                   ':Header><s:Body><getData ' \
                   'xmlns="http://namespace.com"><one>1</one></getData></s' \
                   ':Body></s:Envelope>'


def test_call():
    """Calling a command should result in an http request."""

    s = SoapMessage(
        endpoint='http://endpoint.example.com',
        method='getData',
        parameters=[('one', '1')],
        http_headers={'user-agent': 'sonos'},
        soap_action='ACTION',
        soap_header="<a_header>data</a_header>",
        namespace="http://namespace.com",
        other_arg=4)

    response = mock.MagicMock()
    response.headers = {}
    response.status_code = 200
    response.content = DUMMY_VALID_RESPONSE
    with mock.patch('requests.post', return_value=response) as fake_post:
        result = s.call()
        assert XML.tostring(result)
        fake_post.assert_called_once_with(
            'http://endpoint.example.com',
            headers={'SOAPACTION': '"ACTION"',
                     'Content-Type': 'text/xml; charset="utf-8"', 'user-agent':
                         'sonos'},
            data=mock.ANY, other_arg=4)



# -*- coding: utf-8 -*-
# pylint: disable-msg=too-few-public-methods, redefined-outer-name, no-self-use

"""This file contains the classes used to perform integration tests on the
methods in the SoCo class. They access a real Sonos system.

PLEASE TAKE NOTE: All of these tests are designed to run on a Sonos system
without interfering with normal service. This means that they must not raise
the volume or must leave the player in the same state as they found it in. They
have been made this way since SoCo is developed by volunteers who in all
likelihood do not have a dedicated test system. Accordingly the tests must not
annoy the neighbors, and should return the system to its original state so that
the developers can listen to their music while coding, without having it
interrupted at every unit test!

PLEASE RESPECT THIS.
"""

from __future__ import unicode_literals

import time

import pytest

import pysonos as pysonos_module
from pysonos.data_structures import (
    DidlMusicTrack,
    DidlPlaylistContainer,
    SearchResult,
)
from pysonos.music_library import MusicLibrary
from pysonos.exceptions import SoCoUPnPException

# Mark all tests in this module with the pytest custom "integration" marker so
# they can be selected or deselected as a whole, eg:
# py.test -m "integration"
# or
# py.test -m "no integration"
pytestmark = pytest.mark.integration


@pytest.yield_fixture(scope='session')
def soco(request):
    """Set up and tear down the soco fixture used by all tests."""
    # Get the ip address from the command line, and create the soco object
    # Only one is used per test session, hence the decorator
    ip = request.config.option.IP
    if ip is None:
        pytest.fail("No ip address specified. Use the --ip option.")
    soco = pysonos_module.SoCo(ip)
    # Check the device is playing and has items in the queue
    if len(soco.get_queue()) == 0:
        pytest.fail('Integration tests on the SoCo class must be run '
                    'with at least 1 item in the playlist.')

    transport_info = soco.get_current_transport_info()
    if transport_info['current_transport_state'] != 'PLAYING':
        pytest.fail('Integration tests on the SoCo class must be run '
                    'with the Sonos unit playing.')
    # Save the device's state
    state = {'queue': soco.get_queue(0, 1000),
             'current_track_info': soco.get_current_track_info()}

    # Yield the device to the test function
    yield soco

    # Tear down. Restore state
    soco.stop()
    soco.clear_queue()
    for track in state['queue']:
        soco.add_to_queue(track)
    soco.play_from_queue(
        int(state['current_track_info']['playlist_position']) - 1)
    soco.seek(state['current_track_info']['position'])
    soco.play()


def wait(interval=0.1):
    """Convenience function to adjust sleep interval for all tests."""
    time.sleep(interval)


class TestVolume(object):
    """Integration tests for the volume property."""

    valid_values = range(101)

    @pytest.yield_fixture(autouse=True)
    def restore_volume(self, soco):
        """A fixture which restores volume after each test in the class is
        run."""
        old = soco.volume
        yield
        soco.volume = old
        wait()

    def test_get_and_set(self, soco):
        """Test if the set functionlity works when given valid arguments."""
        old = soco.volume
        assert old in self.valid_values
        if old == self.valid_values[0]:
            new = old + 1
        else:
            new = old - 1
        soco.volume = new
        wait()
        assert soco.volume == new

    def test_invalid_arguments(self, soco):
        """Test if the set functionality coerces into range when given integers
        outside of allowed range."""

        # NOTE We don't test coerce from too large values, since that would
        # put the unit at full volume
        soco.volume = self.valid_values[0] - 1
        wait()
        assert soco.volume == 0

    def test_set_0(self):
        """ Test whether the volume can be set to 0. Regression test for:
        https://github.com/rahims/soco/issues/29
        """

        soco.volume = 0
        wait()
        assert soco.volume == 0


class TestBass(object):
    """Integration tests for the bass property.

    This class implements a full boundary value test.
    """

    valid_values = range(-10, 11)

    @pytest.yield_fixture(autouse=True)
    def restore_bass(self, soco):
        """A fixture which restores bass EQ after each test in the class is
        run."""
        old = soco.bass
        yield
        soco.bass = old
        wait()

    def test_get_and_set(self, soco):
        """Test if the set functionlity works when given valid arguments."""
        assert soco.bass in self.valid_values
        # Values on the boundaries of the valid equivalence partition
        for value in [self.valid_values[0], self.valid_values[-1]]:
            soco.bass = value
            wait()
            assert soco.bass == value

    def test_invalid_arguments(self, soco):
        """Test if the set functionality produces the expected "coerce in
        range" functionality when given a value outside of its range."""
        # Values on the boundaries of the two invalid equivalence partitions
        soco.bass = self.valid_values[0] - 1
        wait()
        assert soco.bass == self.valid_values[0]
        soco.bass = self.valid_values[-1] + 1
        wait()
        assert soco.bass == self.valid_values[-1]


class TestTreble(object):
    """Integration tests for the treble property.

    This class implements a full boundary value test.
    """

    valid_values = range(-10, 11)

    @pytest.yield_fixture(autouse=True)
    def restore_treble(self, soco):
        """A fixture which restores treble EQ after each test in the class is
        run."""
        old = soco.treble
        yield
        soco.treble = old
        wait()

    def test_get_and_set(self, soco):
        """Test if the set functionlity works when given valid arguments."""
        assert soco.treble in self.valid_values
        # Values on the boundaries of the valid equivalence partition
        for value in [self.valid_values[0], self.valid_values[-1]]:
            soco.treble = value
            wait()
            assert soco.treble == value

    def test_invalid_arguments(self, soco):
        """Test if the set functionality produces the expected "coerce in
        range" functionality when given a value outside its range."""
        # Values on the boundaries of the two invalid equivalence partitions
        soco.treble = self.valid_values[0] - 1
        wait()
        assert soco.treble == self.valid_values[0]
        soco.treble = self.valid_values[-1] + 1
        wait()
        assert soco.treble == self.valid_values[-1]


class TestMute(object):
    """Integration test for the mute method."""

    def test(self, soco):
        """Test if the mute method works."""
        old = soco.mute
        assert old is False, ('The unit should not be muted when running '
                              'the unit tests.')
        soco.mute = True
        wait()
        new = soco.mute
        assert new is True
        soco.mute = False
        wait()
        assert soco.mute is False


class TestGetCurrentTransportInfo(object):
    """Integration test for the get_current_transport_info method."""

    # The values in this list must be kept up to date with the values in
    # the test doc string
    transport_info_keys = sorted(['current_transport_status',
                                  'current_transport_state',
                                  'current_transport_speed'])

    def test(self, soco):
        """ Test if the return value is a dictionary that contains the keys:
        current_transport_status, current_transport_state,
        current_transport_speed and that values have been found for all keys,
        i.e. they are not None.
        """
        transport_info = soco.get_current_transport_info()
        assert isinstance(transport_info, dict)
        assert self.transport_info_keys == sorted(transport_info.keys())
        for _, value in transport_info.items():
            assert value is not None


class TestTransport(object):
    """Integration tests for transport methods (play, pause etc)."""

    def test_pause_and_play(self, soco):
        """Test if the pause and play methods work."""
        soco.pause()
        wait(1)
        on_pause = soco.get_current_transport_info()['current_transport_state']
        assert on_pause == 'PAUSED_PLAYBACK'
        soco.play()
        wait(1)
        on_play = soco.get_current_transport_info()['current_transport_state']
        assert on_play == 'PLAYING'

    def test_stop(self, soco):
        """Test if the stop method works."""
        soco.stop()
        wait(1)
        new = soco.get_current_transport_info()['current_transport_state']
        assert new == 'STOPPED'
        soco.play()
        wait(1)
        on_play = soco.get_current_transport_info()['current_transport_state']
        assert on_play == 'PLAYING'

    def test_seek_valid(self, soco):
        """Test if the seek method works with valid input."""
        original_position = soco.get_current_track_info()['position']
        # Format 1
        soco.seek('0:00:00')
        wait()
        position = soco.get_current_track_info()['position']
        assert position in ['0:00:00', '0:00:01']
        # Reset and format 2
        soco.seek(original_position)
        soco.seek('00:00:00')
        wait()
        position = soco.get_current_track_info()['position']
        assert position in ['0:00:00', '0:00:01']
        # Clean up
        soco.seek(original_position)
        wait()

    def test_seek_invald(self, soco):
        """Test if the seek method properly fails with invalid input."""
        for string in ['invalid_time_string', '5:12', '6', 'aa:aa:aa']:
            with pytest.raises(ValueError):
                soco.seek(string)


class TestGetCurrentTrackInfo(object):
    """Integration test for the get_current_track_info method."""

    info_keys = sorted(['album', 'artist', 'title', 'uri', 'metadata',
                        'playlist_position', 'duration', 'album_art',
                        'position'])

    def test_get(self, soco):
        """ Test is the return value is a dictinary and contains the following
        keys: album, artist, title, uri, playlist_position, duration,
        album_art and position.
        """
        info = soco.get_current_track_info()
        assert isinstance(info, dict)
        assert sorted(info.keys()) == self.info_keys


class TestGetSpeakerInfo(object):
    """Integration test for the get_speaker_info method."""

    # The values in this list must be kept up to date with the values in
    # the test doc string
    info_keys = sorted(['zone_name', 'zone_icon', 'uid',
                        'serial_number', 'software_version',
                        'hardware_version', 'mac_address'])

    def test(self, soco):
        """ Test if the return value is a dictionary that contains the keys:
        zone_name, zone_icon, uid, serial_number, software_version,
        hardware_version, mac_address
        and that values have been found for all keys, i.e. they are not None.
        """
        speaker_info = soco.get_speaker_info()
        assert isinstance(speaker_info, dict)
        for _, value in speaker_info.items():
            assert value is not None

# TODO: test GetSpeakersIp


class TestGetQueue(object):
    """Integration test for the get_queue method."""

    # The values in this list must be kept up to date with the values in
    # the test doc string
    queue_element_keys = sorted(['album', 'creator', 'resources',
                                 'album_art_uri', 'title'])

    def test_get(self, soco):
        """ Test is return value is a list of DidlMusicTracks and if each of
        the objects contain the attributes: album, creator, resources,
        album_art_uri and title.
        """
        queue = soco.get_queue(0, 100)
        assert isinstance(queue, list)
        for item in queue:
            assert isinstance(item, DidlMusicTrack)
            for key in self.queue_element_keys:
                assert getattr(item, key)


class TestAddToQueue(object):
    """Integration test for the add_to_queue method."""

    def test_add_to_queue(self, soco):
        """Get the current queue, add the last item of the current queue and
        then compare the length of the old queue with the new and check that
        the last two elements are identical."""

        old_queue = soco.get_queue(0, 1000)
        # Add new element and check
        assert (soco.add_to_queue(old_queue[-1])) == len(old_queue) + 1
        wait()
        new_queue = soco.get_queue()
        assert (len(new_queue) - 1) == len(old_queue)
        assert (new_queue[-1].title) == (new_queue[-2].title)


class TestRemoveFromQueue(object):
    """Integration test for the remove_from_queue method."""

    def test(self, soco):
        """Test if the remove_from_queue method works."""
        old_queue = soco.get_queue()
        soco.remove_from_queue(len(old_queue) - 1)  # queue index is 0 based
        wait()
        new_queue = soco.get_queue()
        assert old_queue != new_queue, (
            'No difference between '
            'queues before and after removing the last item')
        assert len(new_queue) == len(old_queue) - 1


class TestSonosPlaylist(object):
    """Integration tests for Sonos Playlist Management."""

    existing_playlists = None
    playlist_name = 'zSocoTestPlayList42'

    @pytest.yield_fixture(autouse=True)
    def restore_sonos_playlists(self, soco):
        """A fixture which cleans up after each sonos playlist test."""
        if self.existing_playlists is None:
            self.existing_playlists = soco.get_sonos_playlists()
            if self.playlist_name in [x.title
                                      for x in self.existing_playlists]:
                msg = '%s is an existing playlist.' % self.playlist_name
                pytest.fail(msg)

        yield
        for sonos_playlist in soco.get_sonos_playlists():
            if sonos_playlist.title == self.playlist_name:
                soco.remove_sonos_playlist(sonos_playlist=sonos_playlist)

    def test_create(self, soco):
        """Test creating a new empty Sonos playlist."""
        existing_playlists = {x.item_id for x in soco.get_sonos_playlists()}
        new_playlist = soco.create_sonos_playlist(title=self.playlist_name)
        assert type(new_playlist) is DidlPlaylistContainer

        new_pl = {x.item_id for x in soco.get_sonos_playlists()}
        assert new_pl != existing_playlists
        assert new_pl - existing_playlists == {new_playlist.item_id}

    def test_create_from_queue(self, soco):
        """Test creating a Sonos playlist from the current queue."""
        playlist = soco.create_sonos_playlist_from_queue(self.playlist_name)
        assert type(playlist) is DidlPlaylistContainer

        prslt = soco.music_library.browse(ml_item=playlist)
        qrslt = soco.get_queue()
        assert len(prslt) == len(qrslt)
        assert prslt.total_matches == qrslt.total_matches
        assert prslt.number_returned == qrslt.number_returned
        # compare uri because item_id is different, SQ:xx/n for playlist
        for p_item, q_item in zip(prslt, qrslt):
            assert p_item.resources[0].uri == q_item.resources[0].uri

    def test_remove_playlist(self, soco):
        """Test removing a Sonos playlist."""
        # a place holder, remove_sonos_playlist is exercised in the
        # 'restore_sonos_playlists'
        pass

    def test_remove_playlist_itemid(self, soco):
        """Test removing a Sonos playlist by item_id."""
        new_playlist = soco.create_sonos_playlist(title=self.playlist_name)
        assert type(new_playlist) is DidlPlaylistContainer
        assert soco.remove_sonos_playlist(new_playlist.item_id)
        found = False
        for sonos_playlist in soco.get_sonos_playlists():
            if sonos_playlist.title == self.playlist_name:
                found = True
                break
        assert found is False, "new_playlist was not removed by item_id"

    def test_remove_playlist_bad_id(self, soco):
        """Test attempting to remove a Sonos playlist using a bad id."""
        # junky bad
        with pytest.raises(SoCoUPnPException):
            soco.remove_sonos_playlist('SQ:-7')
        # realistic non-existing
        hpl_i = max([int(x.item_id.split(':')[1])
                     for x in soco.get_sonos_playlists()])
        with pytest.raises(SoCoUPnPException):
            soco.remove_sonos_playlist('SQ:{}'.format(hpl_i + 1))


class TestTimer(object):
    """Integration tests for timers on Sonos"""

    existing_timer = None

    @pytest.yield_fixture(autouse=True)
    def restore_timer(self, soco):
        """A fixture which cleans up after each timer test."""
        existing_timer = soco.get_sleep_timer()

        yield
        soco.set_sleep_timer(existing_timer)

    def test_get_set_timer(self, soco):
        """Test setting the timer"""
        assert soco.set_sleep_timer(7200) is None
        result = soco.get_sleep_timer()
        if not any(result == s for s in [ 7200, 7199, 7198 ]):
            pytest.fail("Set timer to 7200, but sonos reports back time as %s" % result['RemainingSleepTimerDuration'])

class TestReorderSonosPlaylist(object):
    """Integration tests for Sonos Playlist Management."""
    existing_playlists = None
    playlist_name = 'zSocoTestPlayList42'
    test_playlist = None
    queue_length = None

    @pytest.yield_fixture(autouse=True, scope="class")
    def restore_sonos_playlists(self, soco):
        """A fixture which cleans up after each sonos playlist test."""
        if self.existing_playlists is None:
            self.existing_playlists = soco.get_sonos_playlists()
        if self.playlist_name in [x.title for x in self.existing_playlists]:
            msg = '%s is an existing playlist.' % self.playlist_name
            pytest.fail(msg)

        queue_list = soco.get_queue()
        if len(queue_list) < 2:
            msg = 'You must have 3 or more items in your queue for testing.'
            pytest.fail(msg)
        playlist = soco.create_sonos_playlist_from_queue(self.playlist_name)
        self.__class__.queue_length = soco.queue_size
        self.__class__.test_playlist = playlist
        yield

        soco.contentDirectory.DestroyObject(
            [('ObjectID', self.test_playlist.item_id)]
        )

    def _reset_spl_contents(self, soco):
        """Ensure test playlist matches queue for each test."""
        soco.contentDirectory.DestroyObject(
            [('ObjectID', self.test_playlist.item_id)]
        )
        playlist = soco.create_sonos_playlist_from_queue(self.playlist_name)
        self.__class__.test_playlist = playlist
        return playlist, self.__class__.queue_length

    def test_reverse_track_order(self, soco):
        """Test reversing the tracks in the Sonos playlist."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        tracks = ','.join([str(x) for x in reversed(range(num_tracks))])
        new_pos = ','.join([str(x) for x in range(num_tracks)])
        args = {'sonos_playlist': test_playlist.item_id,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == 0
        assert response['length'] == num_tracks
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        for s_item, q_item in zip(spl, reversed(soco.get_queue())):
            assert s_item.resources[0].uri == q_item.resources[0].uri

    def test_swap_first_two_items(self, soco):
        """Test a use case in doc string. Swapping the positions of the first
            two tracks in the Sonos playlist."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        tracks = [0, ]
        new_pos = [1, ]

        args = {'sonos_playlist': test_playlist.item_id,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == 0
        assert response['length'] == num_tracks
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        que = soco.get_queue()
        assert spl[0].resources[0].uri == que[1].resources[0].uri
        assert spl[1].resources[0].uri == que[0].resources[0].uri
        # FIXME remove the list on spl and que before slicing, when
        # the deprecated __getitem__ on ListOfMusicInfoItems is
        # removed
        for s_item, q_item in zip(list(spl)[2:], list(que)[2:]):
            assert s_item.resources[0].uri == q_item.resources[0].uri

    def test_remove_first_track(self, soco):
        """Test removing first track from Sonos Playlist."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        tracks = [0, ]
        new_pos = [None, ]

        args = {'sonos_playlist': test_playlist.item_id,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == -1
        assert response['length'] == num_tracks - 1
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        # FIXME remove the list on queue() call, when the deprecated
        # __getitem__ on ListOfMusicInfoItems is removed
        que = list(soco.get_queue())[1:]
        for s_item, q_item in zip(spl, que):
            assert s_item.resources[0].uri == q_item.resources[0].uri

    def test_remove_first_track_full(self, soco):
        """Test removing first track from Sonos Playlist."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        tracks = [0] + list(range(num_tracks - 1))        # [0, 0, 1, ..., n-1]
        new_pos = [None, ] + list(range(num_tracks - 1))  # [None, 0, ..., n-1]
        args = {'sonos_playlist': test_playlist.item_id,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == -1
        assert response['length'] == num_tracks - 1
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        # FIXME remove the list on queue() call, when the deprecated
        # __getitem__ on ListOfMusicInfoItems is removed
        que = list(soco.get_queue())[1:]
        for s_item, q_item in zip(spl, que):
            assert s_item.resources[0].uri == q_item.resources[0].uri

    def test_remove_last_track(self, soco):
        """Test removing last track from Sonos Playlist."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        tracks = range(num_tracks)
        new_pos = list(range(num_tracks - 1)) + [None, ]
        args = {'sonos_playlist': test_playlist.item_id,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == -1
        assert response['length'] == num_tracks - 1
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        # FIXME remove the list on queue() call, when the deprecated
        # __getitem__ on ListOfMusicInfoItems is removed
        que = list(soco.get_queue())[:-1]
        for s_item, q_item in zip(spl, que):
            assert s_item.resources[0].uri == q_item.resources[0].uri

    def test_remove_between_track(self, soco):
        """Test removing a middle track from Sonos Playlist."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        ndx = int(num_tracks / 2)
        tracks = [ndx]
        new_pos = [None]
        args = {'sonos_playlist': test_playlist.item_id,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == -1
        assert response['length'] == num_tracks - 1
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        que = soco.get_queue()
        del que[ndx]
        for s_item, q_item in zip(spl, que):
            assert s_item.resources[0].uri == q_item.resources[0].uri

    def test_remove_some_tracks(self, soco):    # pylint: disable=R0914
        """Test removing some tracks from Sonos Playlist."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        # get rid of the even numbered tracks
        tracks = sorted([x for x in range(num_tracks) if not x & 1],
                        reverse=True)
        new_pos = [None for _ in tracks]
        args = {'sonos_playlist': test_playlist.item_id,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == -1 * len(new_pos)
        assert response['length'] == num_tracks + response['change']
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        que = soco.get_queue()
        for ndx in tracks:
            del que[ndx]
        for s_item, q_item in zip(spl, que):
            assert s_item.resources[0].uri == q_item.resources[0].uri

    def test_remove_all_tracks(self, soco):
        """Test removing all tracks from Sonos Playlist."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        # get rid of the even numbered tracks
        tracks = sorted(range(num_tracks), reverse=True)
        new_pos = [None for _ in tracks]
        args = {'sonos_playlist': test_playlist.item_id,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == -1 * num_tracks
        assert response['length'] == num_tracks + response['change']
        assert response['length'] == 0
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        assert len(spl) == 0

    def test_reorder_and_remove_track(self, soco):
        """Test reorder and removing a track from Sonos Playlist."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        tracks = [1, 2]
        new_pos = [0, None]
        args = {'sonos_playlist': test_playlist.item_id,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == -1
        assert response['length'] == num_tracks + response['change']
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        que = soco.get_queue()
        assert spl[0].resources[0].uri == que[1].resources[0].uri

    def test_object_id_is_object(self, soco):
        """Test removing all tracks from Sonos Playlist."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        tracks = sorted(range(num_tracks), reverse=True)
        new_pos = [None for _ in tracks]
        args = {'sonos_playlist': test_playlist,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == -1 * num_tracks
        assert response['length'] == num_tracks + response['change']
        assert response['length'] == 0
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        assert len(spl) == 0

    def test_remove_all_string(self, soco):
        """Remove all in one op by using strings."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        # we know what we are doing
        tracks = ','.join([str(x) for x in range(num_tracks)])
        new_pos = ''
        args = {'sonos_playlist': test_playlist,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == -1 * num_tracks
        assert response['length'] == num_tracks + response['change']
        assert response['length'] == 0
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        assert len(spl) == 0

    def test_remove_and_reorder_string(self, soco):
        """test remove then reorder using string arguments."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        tracks = '0,2'      # trackA, trackB, trackC, ...
        new_pos = ',0'      # trackC, trackB, ...
        args = {'sonos_playlist': test_playlist,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == -1
        assert response['length'] == num_tracks + response['change']
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        que = soco.get_queue()
        assert spl[0].resources[0].uri == que[2].resources[0].uri
        assert spl[1].resources[0].uri == que[1].resources[0].uri

    def test_move_track_string(self, soco):
        """Test a simple move with strings."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        tracks = "0"
        new_pos = "1"
        args = {'sonos_playlist': test_playlist.item_id,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == 0
        assert response['length'] == num_tracks
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        que = soco.get_queue()
        assert spl[0].resources[0].uri == que[1].resources[0].uri
        assert spl[1].resources[0].uri == que[0].resources[0].uri
        # FIXME remove the list on spl and que before slicing, when
        # the deprecated __getitem__ on ListOfMusicInfoItems is
        # removed
        for s_item, q_item in zip(list(spl)[2:], list(que)[2:]):
            assert s_item.resources[0].uri == q_item.resources[0].uri

    def test_move_track_int(self, soco):
        """Test a simple move with ints."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        tracks = 1
        new_pos = 0
        args = {'sonos_playlist': test_playlist.item_id,
                'tracks': tracks,
                'new_pos': new_pos}
        response = soco.reorder_sonos_playlist(**args)
        assert response['change'] == 0
        assert response['length'] == num_tracks
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        que = soco.get_queue()
        assert spl[0].resources[0].uri == que[1].resources[0].uri
        assert spl[1].resources[0].uri == que[0].resources[0].uri
        # FIXME remove the list on spl and que before slicing, when
        # the deprecated __getitem__ on ListOfMusicInfoItems is
        # removed
        for s_item, q_item in zip(list(spl)[2:], list(que)[2:]):
            assert s_item.resources[0].uri == q_item.resources[0].uri

    def test_clear_sonos_playlist(self, soco):
        """Test the clear_sonos_playlist helper function."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        response = soco.clear_sonos_playlist(test_playlist)
        assert response['change'] == -1 * num_tracks
        assert response['length'] == num_tracks + response['change']
        assert response['length'] == 0
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        assert len(spl) == 0

    def test_clear_empty_sonos_playlist(self, soco):
        """Test clearing an already empty Sonos playlist."""
        test_playlist, _ = self._reset_spl_contents(soco)
        response = soco.clear_sonos_playlist(test_playlist)
        assert response['length'] == 0
        update_id = response['update_id']
        new_response = soco.clear_sonos_playlist(test_playlist,
                                                 update_id=update_id)
        assert new_response['change'] == 0
        assert new_response['length'] == 0
        assert new_response['update_id'] == update_id

    def test_move_in_sonos_playlist(self, soco):
        """Test method move_in_sonos_playlist."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        args = {'sonos_playlist': test_playlist.item_id,
                'track': 0,
                'new_pos': 1}
        response = soco.move_in_sonos_playlist(**args)
        assert response['change'] == 0
        assert response['length'] == num_tracks
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        que = soco.get_queue()
        assert spl[0].resources[0].uri == que[1].resources[0].uri
        assert spl[1].resources[0].uri == que[0].resources[0].uri
        # FIXME remove the list on spl and que before slicing, when
        # the deprecated __getitem__ on ListOfMusicInfoItems is
        # removed
        for s_item, q_item in zip(list(spl)[2:], list(que)[2:]):
            assert s_item.resources[0].uri == q_item.resources[0].uri

    def test_remove_from_sonos_playlist(self, soco):
        """Test remove_from_sonos_playlist method."""
        test_playlist, num_tracks = self._reset_spl_contents(soco)
        args = {'sonos_playlist': test_playlist.item_id,
                'track': 0}
        response = soco.remove_from_sonos_playlist(**args)
        assert response['change'] == -1
        assert response['length'] == num_tracks - 1
        assert response['update_id'] != 0
        spl = soco.music_library.browse(ml_item=test_playlist)
        # FIXME remove the list on queue() call, when the deprecated
        # __getitem__ on ListOfMusicInfoItems is removed
        que = list(soco.get_queue())[1:]
        for s_item, q_item in zip(spl, que):
            assert s_item.resources[0].uri == q_item.resources[0].uri

    def test_get_sonos_playlist_by_attr(self, soco):
        """Test test_get_sonos_playlist_by_attr."""
        test_playlist, _ = self._reset_spl_contents(soco)
        by_name = soco.get_sonos_playlist_by_attr('title', self.playlist_name)
        assert test_playlist.item_id == by_name.item_id
        by_id = soco.get_sonos_playlist_by_attr('item_id',
                                                test_playlist.item_id)
        assert test_playlist.item_id == by_id.item_id
        with pytest.raises(AttributeError):
            soco.get_sonos_playlist_by_attr('fred', 'wilma')

        with pytest.raises(ValueError):
            soco.get_sonos_playlist_by_attr('item_id', 'wilma')


class TestMusicLibrary(object):
    """The the music library methods"""

    search_types = list(MusicLibrary.SEARCH_TRANSLATION.keys())
    specific_search_methods = (
        "artists", "album_artists", "albums", "genres", "composers", "tracks",
        "playlists", "sonos_favorites", "favorite_radio_stations",
        "favorite_radio_shows",
    )

    @pytest.mark.parametrize("search_type", specific_search_methods)
    def test_from_specific_search_methods(self, soco, search_type):
        """Test getting favorites from the music library"""
        search_method = getattr(soco.music_library, "get_" + search_type)
        search_result = search_method()
        assert isinstance(search_result, SearchResult)

    @pytest.mark.parametrize("search_type", search_types)
    def test_music_library_information(self, soco, search_type):
        """Test getting favorites from the music library"""
        search_result = soco.music_library.get_music_library_information(search_type)
        assert isinstance(search_result, SearchResult)



# -*- coding: utf-8 -*-
"""Tests for the services module."""

# These tests require pytest, and mock. Mock comes with Python 3.3, but has
# also been backported for Python 2.7. It is available on pypi.

from __future__ import unicode_literals

import pytest

from pysonos.exceptions import SoCoUPnPException
from pysonos.services import (
    Service, Action, Argument, Vartype
)

try:
    from unittest import mock
except:
    import mock  # TODO: add mock to requirements

# Dummy known-good errors/responses etc.  These are not necessarily valid as
# actual commands, but are valid XML/UPnP. They also contain unicode characters
# to test unicode handling.

DUMMY_ERROR = "".join([
    '<?xml version="1.0"?>',
    '<s:Envelope ',
    'xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" ',
    's:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">',
        '<s:Body>',
            '<s:Fault>',
                '<faultcode>s:Client</faultcode>',
                '<faultstring>UPnPError</faultstring>',
                '<detail>',
                    '<UPnPError xmlns="urn:schemas-upnp-org:control-1-0">',
                        '<errorCode>607</errorCode>',
                        '<errorDescription>Oops μИⅠℂ☺ΔЄ💋</errorDescription>',
                    '</UPnPError>',
                '</detail>',
            '</s:Fault>',
        '</s:Body>',
    '</s:Envelope>'])  # noqa PEP8

DUMMY_VALID_RESPONSE = "".join([
    '<?xml version="1.0"?>',
    '<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"',
        ' s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">',
        '<s:Body>',
            '<u:GetLEDStateResponse ',
                'xmlns:u="urn:schemas-upnp-org:service:DeviceProperties:1">',
                '<CurrentLEDState>On</CurrentLEDState>',
                '<Unicode>μИⅠℂ☺ΔЄ💋</Unicode>',
        '</u:GetLEDStateResponse>',
        '</s:Body>',
    '</s:Envelope>'])  # noqa PEP8

DUMMY_VALID_ACTION = "".join([
    '<?xml version="1.0"?>',
    '<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/"',
        ' s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">',
        '<s:Body>',
            '<u:SetAVTransportURI ',
                'xmlns:u="urn:schemas-upnp-org:service:Service:1">',
                '<InstanceID>0</InstanceID>',
                '<CurrentURI>URI</CurrentURI>',
                '<CurrentURIMetaData></CurrentURIMetaData>',
                '<Unicode>μИⅠℂ☺ΔЄ💋</Unicode>'
            '</u:SetAVTransportURI>',
        '</s:Body>'
    '</s:Envelope>'])  # noqa PEP8

DUMMY_VARTYPE = Vartype('string', None, None, None)

DUMMY_ACTIONS = [
    Action(
        name='Test',
        in_args=[
            Argument(name='Argument1', vartype=DUMMY_VARTYPE),
            Argument(name='Argument2', vartype=DUMMY_VARTYPE)
        ],
        out_args=[]
    )
]

DUMMY_ARGS = [('Argument1', 1), ('Argument2', 2)]

DUMMY_ARGS_ALTERNATIVE = [('Argument1', 3), ('Argument2', 2)]

@pytest.fixture()
def service():
    """A mock Service, for use as a test fixture."""

    mock_soco = mock.MagicMock()
    mock_soco.ip_address = "192.168.1.101"
    mock_service = Service(mock_soco)
    return mock_service


def test_init_defaults(service):
    """Check default properties are set up correctly."""
    assert service.service_type == "Service"
    assert service.version == 1
    assert service.service_id == "Service"
    assert service.base_url == "http://192.168.1.101:1400"
    assert service.control_url == "/Service/Control"
    assert service.scpd_url == "/xml/Service1.xml"
    assert service.event_subscription_url == "/Service/Event"


def test_method_dispatcher_function_creation(service):
    """Testing __getattr__ functionality."""
    import inspect
    # There should be no testing method
    assert 'testing' not in service.__dict__.keys()
    # but we should be able to inspect it
    assert inspect.ismethod(service.testing)
    # and then, having examined it, the method should be cached on the instance
    assert 'testing' in service.__dict__.keys()
    assert service.testing.__name__ == "testing"
    # check that send_command is actually called when we invoke a method
    service.send_command = lambda x, y: "Hello {}".format(x)
    assert service.testing(service) == "Hello testing"


def test_method_dispatcher_arg_count(service):
    """_dispatcher should pass its args to send_command."""
    service.send_command = mock.Mock()
    # http://bugs.python.org/issue7688
    # __name__ must be a string in python 2
    method = service.__getattr__(str('test'))
    assert method('onearg')
    service.send_command.assert_called_with('test', 'onearg')
    assert method()  # no args
    service.send_command.assert_called_with('test')
    assert method('one', cache_timeout=4)  # one arg + cache_timeout
    service.send_command.assert_called_with('test', 'one', cache_timeout=4)


def test_wrap(service):
    """wrapping args in XML properly."""
    assert service.wrap_arguments([('first', 'one'), ('second', 2)]) == \
        "<first>one</first><second>2</second>"
    assert service.wrap_arguments() == ""
    # Unicode
    assert service.wrap_arguments([('unicode', "μИⅠℂ☺ΔЄ💋")]) == \
        "<unicode>μИⅠℂ☺ΔЄ💋</unicode>"
    # XML escaping - do we also need &apos; ?
    assert service.wrap_arguments([('weird', '&<"2')]) == \
        "<weird>&amp;&lt;&quot;2</weird>"


def test_unwrap(service):
    """unwrapping args from XML."""
    assert service.unwrap_arguments(DUMMY_VALID_RESPONSE) == {
        "CurrentLEDState": "On",
        "Unicode": "μИⅠℂ☺ΔЄ💋"}


def test_unwrap_invalid_char(service):
    """Test unwrapping args from XML with invalid char"""
    responce_with_invalid_char = DUMMY_VALID_RESPONSE.replace(
        'μИⅠℂ☺ΔЄ💋', 'AB'
    )
    # Note, the invalid ^D (code point 0x04) should be filtered out
    assert service.unwrap_arguments(responce_with_invalid_char) == {
        "CurrentLEDState": "On",
        "Unicode": "AB"}


def test_compose(service):
    """Test argument composition."""
    service._actions = DUMMY_ACTIONS
    service.DEFAULT_ARGS = {}

    # Detect unknown action
    with pytest.raises(AttributeError):
        service.compose_args('Error', {})
    # Detect missing / unknown arguments
    with pytest.raises(ValueError):
        service.compose_args('Test', {'Argument1': 1})
    with pytest.raises(ValueError):
        service.compose_args('Test', dict(DUMMY_ARGS+[('Error', 3)]))

    # Check correct output
    assert service.compose_args('Test', dict(DUMMY_ARGS)) == DUMMY_ARGS

    # Set Argument1 = 1 as default
    service.DEFAULT_ARGS = dict(DUMMY_ARGS[:1])

    # Check that arguments are completed with default values
    assert service.compose_args('Test', dict(DUMMY_ARGS[1:])) == DUMMY_ARGS
    # Check that given arguments override the default values
    assert service.compose_args('Test', dict(DUMMY_ARGS_ALTERNATIVE)) \
        == DUMMY_ARGS_ALTERNATIVE



def test_build_command(service):
    """Test creation of SOAP body and headers from a command."""
    headers, body = service.build_command('SetAVTransportURI', [
        ('InstanceID', 0),
        ('CurrentURI', 'URI'),
        ('CurrentURIMetaData', ''),
        ('Unicode', 'μИⅠℂ☺ΔЄ💋')
    ])
    assert body == DUMMY_VALID_ACTION
    assert headers == {
        'Content-Type': 'text/xml; charset="utf-8"',
        'SOAPACTION':
        'urn:schemas-upnp-org:service:Service:1#SetAVTransportURI'}


def test_send_command(service):
    """Calling a command should result in a http request, unless the cache is
    hit."""
    response = mock.MagicMock()
    response.headers = {}
    response.status_code = 200
    response.text = DUMMY_VALID_RESPONSE
    with mock.patch('requests.post', return_value=response) as fake_post:
        result = service.send_command('SetAVTransportURI', [
            ('InstanceID', 0),
            ('CurrentURI', 'URI'),
            ('CurrentURIMetaData', ''),
            ('Unicode', 'μИⅠℂ☺ΔЄ💋')
        ], cache_timeout=2)
        assert result == {'CurrentLEDState': 'On', 'Unicode': "μИⅠℂ☺ΔЄ💋"}
        fake_post.assert_called_once_with(
            'http://192.168.1.101:1400/Service/Control',
            headers=mock.ANY, data=DUMMY_VALID_ACTION.encode('utf-8'), timeout=20)
        # Now the cache should be primed, so try it again
        fake_post.reset_mock()
        result = service.send_command('SetAVTransportURI', [
            ('InstanceID', 0),
            ('CurrentURI', 'URI'),
            ('CurrentURIMetaData', ''),
            ('Unicode', 'μИⅠℂ☺ΔЄ💋')
        ], cache_timeout=0)
        # The cache should be hit, so there should be no http request
        assert not fake_post.called
        # but this should not affefct a call with different params
        fake_post.reset_mock()
        result = service.send_command('SetAVTransportURI', [
            ('InstanceID', 1),
            ('CurrentURI', 'URI2'),
            ('CurrentURIMetaData', 'abcd'),
            ('Unicode', 'μИⅠℂ☺ΔЄ💋')
        ])
        assert fake_post.called
        # calling again after the time interval will avoid the cache
        fake_post.reset_mock()
        import time
        time.sleep(2)
        result = service.send_command('SetAVTransportURI', [
            ('InstanceID', 0),
            ('CurrentURI', 'URI'),
            ('CurrentURIMetaData', ''),
            ('Unicode', 'μИⅠℂ☺ΔЄ💋')
        ])
        assert fake_post.called


def test_handle_upnp_error(service):
    """Check errors are extracted properly."""
    with pytest.raises(SoCoUPnPException) as E:
        service.handle_upnp_error(DUMMY_ERROR)
    assert "UPnP Error 607 received: Signature Failure from 192.168.1.101" \
        == E.value.message
    assert E.value.error_code == '607'
    assert E.value.error_description == 'Signature Failure'
    # TODO: Try this with a None Error Code

# TODO: test iter_actions



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import mock
import pytest

from pysonos import SoCo
from pysonos.data_structures import (
    DidlMusicTrack, to_didl_string
)
from pysonos.exceptions import (
    SoCoSlaveException, SoCoUPnPException
)
from pysonos.groups import ZoneGroup
from pysonos.xml import XML

IP_ADDR = '192.168.1.101'


@pytest.yield_fixture()
def moco():
    """A mock soco with fake services and hardcoded is_coordinator.

    Allows calls to services to be tracked. Should not cause any network
    access
    """
    services = (
        'AVTransport', 'RenderingControl', 'DeviceProperties',
        'ContentDirectory', 'ZoneGroupTopology')
    patchers = [mock.patch('pysonos.core.{}'.format(service))
                for service in services]
    for patch in patchers:
        patch.start()
    with mock.patch("pysonos.SoCo.is_coordinator",
                    new_callable=mock.PropertyMock) as is_coord:
        is_coord = True
        yield SoCo(IP_ADDR)
    for patch in reversed(patchers):
        patch.stop()


@pytest.yield_fixture()
def moco_only_on_master():
    """A mock soco with fake services.

    Allows calls to services to be tracked. Should not cause any network
    access
    """
    services = (
        'AVTransport', 'RenderingControl', 'DeviceProperties',
        'ContentDirectory', 'ZoneGroupTopology')
    patchers = [mock.patch('pysonos.core.{}'.format(service))
                for service in services]
    for patch in patchers:
        patch.start()
    yield SoCo(IP_ADDR)
    for patch in reversed(patchers):
        patch.stop()


ZGS = """<ZoneGroupState>
      <ZoneGroups>
        <ZoneGroup Coordinator="RINCON_000XXX1400" ID="RINCON_000E5876A0E801400:210">
          <ZoneGroupMember
              UUID="RINCON_000XXX1400"
              Location="http://192.168.1.101:1400/xml/device_description.xml"
              ZoneName="Kitchen"
              Icon="x-rincon-roomicon:kitchen"
              Configuration="1"
              SoftwareVersion="49.2-64250"
              MinCompatibleVersion="48.0-00000"
              LegacyCompatibleVersion="36.0-00000"
              BootSeq="162"
              TVConfigurationError="0"
              HdmiCecAvailable="0"
              WirelessMode="0"
              WirelessLeafOnly="0"
              HasConfiguredSSID="1"
              ChannelFreq="2437"
              BehindWifiExtender="0"
              WifiEnabled="1"
              Orientation="0"
              RoomCalibrationState="4"
              SecureRegState="3"
              VoiceConfigState="0"
              MicEnabled="0"
              AirPlayEnabled="0"
              IdleState="1"
              MoreInfo=""/>
          <ZoneGroupMember
              UUID="RINCON_B8E93781F3EA01400"
              Location="http://192.168.1.31:1400/xml/device_description.xml"
              ZoneName="Living Room"
              Icon="x-rincon-roomicon:masterbedroom"
              Configuration="1"
              SoftwareVersion="49.2-64250"
              MinCompatibleVersion="48.0-00000"
              LegacyCompatibleVersion="36.0-00000"
              BootSeq="226"
              TVConfigurationError="0"
              HdmiCecAvailable="0"
              WirelessMode="0"
              WirelessLeafOnly="0"
              HasConfiguredSSID="1"
              ChannelFreq="2437"
              BehindWifiExtender="0"
              WifiEnabled="1"
              Orientation="0"
              RoomCalibrationState="4"
              SecureRegState="3"
              VoiceConfigState="0"
              MicEnabled="0"
              AirPlayEnabled="0"
              IdleState="1"
              MoreInfo=""/>
        </ZoneGroup>
        <ZoneGroup Coordinator="RINCON_000E58A53FAE01400" ID="RINCON_000E58A53FAE01400:107">
          <ZoneGroupMember
              UUID="RINCON_000E58A53FAE01400"
              Location="http://192.168.1.173:1400/xml/device_description.xml"
              ZoneName="Stue"
              Icon="x-rincon-roomicon:living"
              Configuration="1"
              SoftwareVersion="49.2-64250"
              MinCompatibleVersion="48.0-00000"
              LegacyCompatibleVersion="36.0-00000"
              BootSeq="1777"
              TVConfigurationError="0"
              HdmiCecAvailable="0"
              WirelessMode="0"
              WirelessLeafOnly="0"
              HasConfiguredSSID="0"
              ChannelFreq="2437"
              BehindWifiExtender="0"
              WifiEnabled="1"
              Orientation="0"
              RoomCalibrationState="4"
              SecureRegState="3"
              VoiceConfigState="0"
              MicEnabled="0"
              AirPlayEnabled="0"
              IdleState="1"
              MoreInfo=""/>
        </ZoneGroup>
        <ZoneGroup Coordinator="RINCON_000E5884455C01400" ID="RINCON_000E5884455C01400:114">
          <ZoneGroupMember
              UUID="RINCON_000E5884455C01400"
              Location="http://192.168.1.197:1400/xml/device_description.xml"
              ZoneName="Kontor"
              Icon="x-rincon-roomicon:office"
              Configuration="1"
              SoftwareVersion="49.2-64250"
              MinCompatibleVersion="48.0-00000"
              LegacyCompatibleVersion="36.0-00000"
              BootSeq="196"
              TVConfigurationError="0"
              HdmiCecAvailable="0"
              WirelessMode="0"
              WirelessLeafOnly="0"
              HasConfiguredSSID="1"
              ChannelFreq="2437"
              BehindWifiExtender="0"
              WifiEnabled="1"
              Orientation="0"
              RoomCalibrationState="4"
              SecureRegState="3"
              VoiceConfigState="0"
              MicEnabled="0"
              AirPlayEnabled="0"
              IdleState="0"
              MoreInfo=""/>
        </ZoneGroup>
      </ZoneGroups>
      <VanishedDevices></VanishedDevices>
    </ZoneGroupState>"""

@pytest.yield_fixture
def moco_zgs(moco):
    """A mock soco with zone group state."""
    moco.zoneGroupTopology.GetZoneGroupState.return_value = {
        'ZoneGroupState': ZGS
    }
    yield moco


class TestSoco:
    device_description = """<?xml version="1.0" encoding="utf-8" ?>
        <root xmlns="urn:schemas-upnp-org:device-1-0">
          <specVersion>
            <major>1</major>
            <minor>0</minor>
          </specVersion>
          <device>
            <deviceType>urn:schemas-upnp-org:device:ZonePlayer:1</deviceType>
            <friendlyName>""" + IP_ADDR + """ - Sonos Play:5</friendlyName>
            <manufacturer>Sonos, Inc.</manufacturer>
            <manufacturerURL>http://www.sonos.com</manufacturerURL>
            <modelNumber>S3</modelNumber>
            <modelDescription>Sonos Play:5</modelDescription>
            <modelName>Sonos Play:3</modelName>
            <modelURL>http://www.sonos.com/products/zoneplayers/S5</modelURL>
            <softwareVersion>49.2-64250</softwareVersion>
            <hardwareVersion>1.8.1.2-2</hardwareVersion>
            <serialNum>00-11-22-33-44-55:E</serialNum>
            <UDN>uuid:RINCON_000E5884455C01400</UDN>
            <iconList>
              <icon>
                <id>0</id>
                <mimetype>image/png</mimetype>
                <width>48</width>
                <height>48</height>
                <depth>24</depth>
                <url>/img/icon-S3.png</url>
              </icon>
            </iconList>
            <minCompatibleVersion>48.0-00000</minCompatibleVersion>
            <legacyCompatibleVersion>36.0-00000</legacyCompatibleVersion>
            <apiVersion>1.11.1</apiVersion>
            <minApiVersion>1.1.0</minApiVersion>
            <displayVersion>10.1.2</displayVersion>
            <extraVersion>OTP: 1.1.1(1-16-4-zp5s-0.5)</extraVersion>
            <roomName>Room</roomName>
            <displayName>Play:5</displayName>
            <zoneType>5</zoneType>
            <feature1>0x02000002</feature1>
            <feature2>0x00006172</feature2>
            <feature3>0x0003002a</feature3>
            <seriesid>P100</seriesid>
            <variant>0</variant>
            <internalSpeakerSize>3</internalSpeakerSize>
            <bassExtension>0.000</bassExtension>
            <satGainOffset>0.000</satGainOffset>
            <memory>32</memory>
            <flash>32</flash>
            <flashRepartitioned>1</flashRepartitioned>
            <ampOnTime>425</ampOnTime>
            <retailMode>0</retailMode>
            <serviceList>
              <service>
                <serviceType>urn:schemas-upnp-org:service:AlarmClock:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:AlarmClock</serviceId>
                <controlURL>/AlarmClock/Control</controlURL>
                <eventSubURL>/AlarmClock/Event</eventSubURL>
                <SCPDURL>/xml/AlarmClock1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:MusicServices:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:MusicServices</serviceId>
                <controlURL>/MusicServices/Control</controlURL>
                <eventSubURL>/MusicServices/Event</eventSubURL>
                <SCPDURL>/xml/MusicServices1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:AudioIn:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:AudioIn</serviceId>
                <controlURL>/AudioIn/Control</controlURL>
                <eventSubURL>/AudioIn/Event</eventSubURL>
                <SCPDURL>/xml/AudioIn1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:DeviceProperties:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:DeviceProperties</serviceId>
                <controlURL>/DeviceProperties/Control</controlURL>
                <eventSubURL>/DeviceProperties/Event</eventSubURL>
                <SCPDURL>/xml/DeviceProperties1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:SystemProperties:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:SystemProperties</serviceId>
                <controlURL>/SystemProperties/Control</controlURL>
                <eventSubURL>/SystemProperties/Event</eventSubURL>
                <SCPDURL>/xml/SystemProperties1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:ZoneGroupTopology:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:ZoneGroupTopology</serviceId>
                <controlURL>/ZoneGroupTopology/Control</controlURL>
                <eventSubURL>/ZoneGroupTopology/Event</eventSubURL>
                <SCPDURL>/xml/ZoneGroupTopology1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:GroupManagement:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:GroupManagement</serviceId>
                <controlURL>/GroupManagement/Control</controlURL>
                <eventSubURL>/GroupManagement/Event</eventSubURL>
                <SCPDURL>/xml/GroupManagement1.xml</SCPDURL>
              </service>
              <service>
                <serviceType>urn:schemas-tencent-com:service:QPlay:1</serviceType>
                <serviceId>urn:tencent-com:serviceId:QPlay</serviceId>
                <controlURL>/QPlay/Control</controlURL>
                <eventSubURL>/QPlay/Event</eventSubURL>
                <SCPDURL>/xml/QPlay1.xml</SCPDURL>
              </service>
            </serviceList>
            <deviceList>
              <device>
          <deviceType>urn:schemas-upnp-org:device:MediaServer:1</deviceType>
          <friendlyName>192.168.1.197 - Sonos Play:5 Media Server</friendlyName>
          <manufacturer>Sonos, Inc.</manufacturer>
          <manufacturerURL>http://www.sonos.com</manufacturerURL>
          <modelNumber>S5</modelNumber>
          <modelDescription>Sonos Play:5 Media Server</modelDescription>
          <modelName>Sonos Play:5</modelName>
          <modelURL>http://www.sonos.com/products/zoneplayers/S5</modelURL>
          <UDN>uuid:RINCON_000E5884455C01400_MS</UDN>
          <serviceList>
            <service>
              <serviceType>urn:schemas-upnp-org:service:ContentDirectory:1</serviceType>
              <serviceId>urn:upnp-org:serviceId:ContentDirectory</serviceId>
              <controlURL>/MediaServer/ContentDirectory/Control</controlURL>
              <eventSubURL>/MediaServer/ContentDirectory/Event</eventSubURL>
              <SCPDURL>/xml/ContentDirectory1.xml</SCPDURL>
            </service>
            <service>
              <serviceType>urn:schemas-upnp-org:service:ConnectionManager:1</serviceType>
                    <serviceId>urn:upnp-org:serviceId:ConnectionManager</serviceId>
                    <controlURL>/MediaServer/ConnectionManager/Control</controlURL>
                    <eventSubURL>/MediaServer/ConnectionManager/Event</eventSubURL>
                    <SCPDURL>/xml/ConnectionManager1.xml</SCPDURL>
                  </service>
                </serviceList>
              </device>
              <device>
                <deviceType>urn:schemas-upnp-org:device:MediaRenderer:1</deviceType>
          <friendlyName>Kontor - Sonos Play:5 Media Renderer</friendlyName>
          <manufacturer>Sonos, Inc.</manufacturer>
          <manufacturerURL>http://www.sonos.com</manufacturerURL>
          <modelNumber>S5</modelNumber>
          <modelDescription>Sonos Play:5 Media Renderer</modelDescription>
          <modelName>Sonos Play:5</modelName>
          <modelURL>http://www.sonos.com/products/zoneplayers/S5</modelURL>
                <UDN>uuid:RINCON_000E5884455C01400_MR</UDN>
                <serviceList>
                  <service>
                    <serviceType>urn:schemas-upnp-org:service:RenderingControl:1</serviceType>
                    <serviceId>urn:upnp-org:serviceId:RenderingControl</serviceId>
                    <controlURL>/MediaRenderer/RenderingControl/Control</controlURL>
                    <eventSubURL>/MediaRenderer/RenderingControl/Event</eventSubURL>
                    <SCPDURL>/xml/RenderingControl1.xml</SCPDURL>
                  </service>
                  <service>
                    <serviceType>urn:schemas-upnp-org:service:ConnectionManager:1</serviceType>
                    <serviceId>urn:upnp-org:serviceId:ConnectionManager</serviceId>
                    <controlURL>/MediaRenderer/ConnectionManager/Control</controlURL>
                    <eventSubURL>/MediaRenderer/ConnectionManager/Event</eventSubURL>
                    <SCPDURL>/xml/ConnectionManager1.xml</SCPDURL>
                  </service>
                  <service>
                    <serviceType>urn:schemas-upnp-org:service:AVTransport:1</serviceType>
                    <serviceId>urn:upnp-org:serviceId:AVTransport</serviceId>
                    <controlURL>/MediaRenderer/AVTransport/Control</controlURL>
                    <eventSubURL>/MediaRenderer/AVTransport/Event</eventSubURL>
                    <SCPDURL>/xml/AVTransport1.xml</SCPDURL>
                  </service>
                  <service>
                    <serviceType>urn:schemas-sonos-com:service:Queue:1</serviceType>
                    <serviceId>urn:sonos-com:serviceId:Queue</serviceId>
                    <controlURL>/MediaRenderer/Queue/Control</controlURL>
                    <eventSubURL>/MediaRenderer/Queue/Event</eventSubURL>
                    <SCPDURL>/xml/Queue1.xml</SCPDURL>
                  </service>
              <service>
                <serviceType>urn:schemas-upnp-org:service:GroupRenderingControl:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:GroupRenderingControl</serviceId>
                <controlURL>/MediaRenderer/GroupRenderingControl/Control</controlURL>
                <eventSubURL>/MediaRenderer/GroupRenderingControl/Event</eventSubURL>
                <SCPDURL>/xml/GroupRenderingControl1.xml</SCPDURL>
              </service>
                </serviceList>
                <X_Rhapsody-Extension xmlns="http://www.real.com/rhapsody/xmlns/upnp-1-0">
                  <deviceID>urn:rhapsody-real-com:device-id-1-0:sonos_1:RINCON_000E5884455C01400</deviceID>
                    <deviceCapabilities>
                      <interactionPattern type="real-rhapsody-upnp-1-0"/>
                    </deviceCapabilities>
                </X_Rhapsody-Extension>
                <qq:X_QPlay_SoftwareCapability xmlns:qq="http://www.tencent.com">QPlay:2</qq:X_QPlay_SoftwareCapability>
                <iconList>
                  <icon>
                    <mimetype>image/png</mimetype>
                    <width>48</width>
                    <height>48</height>
                    <depth>24</depth>
                    <url>/img/icon-S5.png</url>
                  </icon>
                </iconList>
              </device>
            </deviceList>
          </device>
        </root>
    """

    @pytest.mark.parametrize('bad_ip_addr', ['not_ip', '555.555.555.555'])
    def test_soco_bad_ip(self, bad_ip_addr):
        with pytest.raises(ValueError):
            speaker = SoCo(bad_ip_addr)

    def test_soco_init(self, moco):
        assert moco.ip_address == IP_ADDR
        assert moco.speaker_info == {}

    def test_soco_str(self, moco):
        assert str(moco) == "<SoCo object at ip {}>".format(IP_ADDR)

    def test_soco_repr(self, moco):
        assert repr(moco) == 'SoCo("{}")'.format(IP_ADDR)

    @pytest.mark.parametrize('model_name', (
        ('Play:5', False),
        ('Sonos One', False),
        ('PLAYBAR', True),
        ('Sonos Beam', True),
        ('Sonos Playbar', True),
        ('Sonos Playbase', True)))
    def test_soco_is_soundbar(self, moco, model_name):
        moco._is_soundbar = None
        moco.speaker_info['model_name'] = model_name[0]
        assert moco.is_soundbar == model_name[1]

    @mock.patch("pysonos.core.requests")
    @pytest.mark.parametrize('refresh', [None, False, True])
    def test_soco_get_speaker_info_speaker_not_set_refresh(
            self, mocr, moco_zgs, refresh):
        """Internal speaker_info not set; Refresh all values (default, False,
        True)

        => should update
        """
        response = mock.MagicMock()
        mocr.get.return_value = response
        response.content = self.device_description
        # save old state
        old = moco_zgs.speaker_info
        moco_zgs.speaker_info = {}
        if refresh is None:
            res = moco_zgs.get_speaker_info()
        else:
            res = moco_zgs.get_speaker_info(refresh)
        # restore original value
        moco_zgs.speaker_info = old
        mocr.get.assert_called_once_with(
            'http://' + IP_ADDR + ':1400/xml/device_description.xml',
            timeout=None,
        )
        should = {
            'zone_name': "Room",
            'player_icon': "/img/icon-S3.png",
            'uid': "RINCON_000XXX1400",
            'serial_number': "00-11-22-33-44-55:E",
            'software_version': "49.2-64250",
            'hardware_version': "1.8.1.2-2",
            'model_number': "S3",
            'model_name': "Sonos Play:3",
            'display_version': "10.1.2",
            'mac_address': "00-11-22-33-44-55"
        }
        assert should == res

    @mock.patch("pysonos.core.requests")
    @pytest.mark.parametrize('refresh', [None, False])
    def test_soco_get_speaker_info_speaker_set_no_refresh(
            self, mocr, moco_zgs, refresh):
        """Internal speaker_info set; No refresh (default, False)

        => should not update
        """
        should = {
            'info': "yes"
        }
        # save old state
        old = moco_zgs.speaker_info
        moco_zgs.speaker_info = should
        if refresh is None:
            res = moco_zgs.get_speaker_info()
        else:
            res = moco_zgs.get_speaker_info(refresh)
        # restore original value
        moco_zgs.speaker_info = old
        # got 'should' returned
        assert res is should
        # no network request performed
        assert not mocr.get.called

    @mock.patch("pysonos.core.requests")
    @pytest.mark.parametrize('should', [{}, {'info': "yes"}])
    def test_soco_get_speaker_info_speaker_set_no_refresh(
            self, mocr, moco_zgs, should):
        """Internal speaker_info not set/set; Refresh True.

        => should update
        """
        response = mock.MagicMock()
        mocr.get.return_value = response
        response.content = self.device_description
        # save old state
        old = moco_zgs.speaker_info
        moco_zgs.speaker_info = should
        res = moco_zgs.get_speaker_info(True)
        # restore original value
        moco_zgs.speaker_info = old
        mocr.get.assert_called_once_with(
            'http://' + IP_ADDR + ':1400/xml/device_description.xml',
            timeout=None,
        )
        # get_speaker_info only updates internal speaker_info and does not
        # replace it
        should.update({
            'zone_name': "Room",
            'player_icon': "/img/icon-S3.png",
            'uid': "RINCON_000XXX1400",
            'serial_number': "00-11-22-33-44-55:E",
            'software_version': "29.5-91030",
            'hardware_version': "1.8.1.2-2",
            'model_number': "S3",
            'model_name': "Sonos Play:3",
            'display_version': "10.1.2",
            'mac_address': "00-11-22-33-44-55"
        })
        assert should == res


class TestAVTransport:

    @pytest.mark.parametrize('playmode', [
        "NORMAL", "SHUFFLE_NOREPEAT", "SHUFFLE", "REPEAT_ALL", "SHUFFLE_REPEAT_ONE", "REPEAT_ONE"
    ])
    def test_soco_play_mode_values(self, moco, playmode):
        moco.avTransport.GetTransportSettings.return_value = {
            'PlayMode': playmode}
        assert moco.play_mode == playmode
        moco.avTransport.GetTransportSettings.assert_called_once_with(
            [('InstanceID', 0)]
        )
        moco.avTransport.reset_mock()

    def test_soco_play_mode_bad_value(self, moco):
        with pytest.raises(KeyError):
            moco.play_mode = "BAD_VALUE"
        assert not moco.avTransport.SetPlayMode.called

    def test_soco_play_mode_lowercase(self, moco):
        moco.play_mode = "normal"
        moco.avTransport.SetPlayMode.assert_called_once_with(
            [('InstanceID', 0), ('NewPlayMode', 'NORMAL')]
        )

    def test_soco_cross_fade(self, moco):
        moco.avTransport.GetCrossfadeMode.return_value = {'CrossfadeMode': '1'}
        assert moco.cross_fade
        moco.avTransport.GetCrossfadeMode.return_value = {'CrossfadeMode': '0'}
        assert not moco.cross_fade
        moco.avTransport.GetCrossfadeMode.assert_called_with(
            [('InstanceID', 0)]
        )
        moco.cross_fade = True
        moco.avTransport.SetCrossfadeMode.assert_called_once_with(
            [('InstanceID', 0), ('CrossfadeMode', '1')]
        )

    def test_soco_play(self, moco):
        moco.play()
        moco.avTransport.Play.assert_called_once_with(
            [('InstanceID', 0), ('Speed', 1)]
        )

    # Test that uris are forced to Radio style display and controls when
    # force_radio is True prefix is replaced with "x-rincon-mp3radion://"
    # first set of test with no forcing, second set with force_radio=True

    # No forcing
    @pytest.mark.parametrize("uri_in, uri_passed", [
        ('x-file-cifs://server/MyNiceRing.mp3',
         'x-file-cifs://server/MyNiceRing.mp3'),
        ('http://archive.org/download/TenD2005-07-16t_64kb.mp3',
         'http://archive.org/download/TenD2005-07-16t_64kb.mp3'),
        ('x-sonosapi-radio:station%3a%3aps.147077612?sid=203&flags=76&sn=2',
         'x-sonosapi-radio:station%3a%3aps.147077612?sid=203&flags=76&sn=2'),
    ])
    def test_soco_play_uri(self, moco, uri_in, uri_passed):
        moco.play_uri(uri_in)
        moco.avTransport.SetAVTransportURI.assert_called_once_with([
            ('InstanceID', 0),
            ('CurrentURI', uri_passed),
            ('CurrentURIMetaData', '')
        ])
        moco.avTransport.reset_mock()

    # with force_radio=True
    @pytest.mark.parametrize("uri_in, uri_passed", [
        ('http://archive.org/download/TenD2005-07-16t_64kb.mp3',
         'x-rincon-mp3radio://archive.org/download/TenD2005-07-16t_64kb.mp3'),
        ('https://archive.org/download/TenD2005-07-16t_64kb.mp3',
         'x-rincon-mp3radio://archive.org/download/TenD2005-07-16t_64kb.mp3'),
        ('aac://icy-e-bz-04-cr.sharp-stream.com/magic1054.aac?amsparams=playerid:BMUK_tunein;skey:1483570441&awparams=loggedin:false',
         'x-rincon-mp3radio://icy-e-bz-04-cr.sharp-stream.com/magic1054.aac?amsparams=playerid:BMUK_tunein;skey:1483570441&awparams=loggedin:false')
    ])
    def test_soco_play_uri_force_radio(self, moco, uri_in, uri_passed):
        moco.play_uri(uri_in, force_radio=True)
        moco.avTransport.SetAVTransportURI.assert_called_once_with([
            ('InstanceID', 0),
            ('CurrentURI', uri_passed),
            ('CurrentURIMetaData', '')
        ])
        moco.avTransport.reset_mock()

    def test_soco_play_uri_calls_play(self, moco):
        uri = 'http://archive.org/download/tend2005-07-16.flac16/tend2005-07-16t10wonderboy_64kb.mp3'
        moco.play_uri(uri)

        moco.avTransport.Play.assert_called_with(
            [('InstanceID', 0), ('Speed', 1)]
        )

    def test_soco_play_uri_with_title(self, moco):
        uri = 'http://archive.org/download/tend2005-07-16.flac16/tend2005-07-16t10wonderboy_64kb.mp3'
        moco.play_uri(uri, title='<Fast & Loose>')

        moco.avTransport.SetAVTransportURI.assert_called_with([
            ('InstanceID', 0),
            ('CurrentURI', uri),
            ('CurrentURIMetaData', '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><item id="R:0/0/0" parentID="R:0/0" restricted="true"><dc:title>&lt;Fast &amp; Loose&gt;</dc:title><upnp:class>object.item.audioItem.audioBroadcast</upnp:class><desc id="cdudn" nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">SA_RINCON65031_</desc></item></DIDL-Lite>')
        ])

    def test_soco_pause(self, moco):
        moco.pause()
        moco.avTransport.Pause.assert_called_once_with(
            [('InstanceID', 0), ('Speed', 1)]
        )

    def test_soco_stop(self, moco):
        moco.stop()
        moco.avTransport.Stop.assert_called_once_with(
            [('InstanceID', 0), ('Speed', 1)]
        )

    def test_soco_next(self, moco):
        moco.next()
        moco.avTransport.Next.assert_called_once_with(
            [('InstanceID', 0), ('Speed', 1)]
        )

    def test_soco_previous(self, moco):
        moco.previous()
        moco.avTransport.Previous.assert_called_once_with(
            [('InstanceID', 0), ('Speed', 1)]
        )

    @pytest.mark.parametrize('bad_timestamp', [
        'NOT_VALID',
        '12:34:56:78',
        '99',
        '12:3:4'
    ])
    def test_soco_seek_invalid(self, moco, bad_timestamp):
        with pytest.raises(ValueError):
            moco.seek(bad_timestamp)
        assert not moco.avTransport.Seek.called

    @pytest.mark.parametrize('timestamp', [
        # '12:34', # Should this be valid?
        '1:23:45',
        '10:23:45',
        '12:78:78'  # Should this really be valid?
    ])
    def test_soco_seek_valid(self, moco, timestamp):
        moco.seek(timestamp)
        moco.avTransport.Seek.assert_called_once_with(
            [('InstanceID', 0), ('Unit', 'REL_TIME'),
                ('Target', timestamp)])
        moco.avTransport.reset_mock()

    def test_soco_current_transport_info(self, moco):
        moco.avTransport.GetTransportInfo.return_value = {
            'CurrentTransportState': 'PLAYING',
            'CurrentTransportStatus': 'OK',
            'CurrentSpeed': '1'}
        playstate = moco.get_current_transport_info()
        moco.avTransport.GetTransportInfo.assert_called_once_with(
            [('InstanceID', 0)]
        )
        assert playstate['current_transport_state'] == 'PLAYING'
        assert playstate['current_transport_status'] == 'OK'
        assert playstate['current_transport_speed'] == '1'

    def test_soco_get_queue(self, moco):
        moco.contentDirectory.Browse.return_value = {
            'Result': '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><item id="Q:0/1" parentID="Q:0" restricted="true"><res protocolInfo="fake.com-fake-direct:*:audio/mp3:*" duration="0:02:32">radea:Tra.12345678.mp3</res><upnp:albumArtURI>/getaa?r=1&amp;u=radea%3aTra.12345678.mp3</upnp:albumArtURI><dc:title>Item 1 Title</dc:title><upnp:class>object.item.audioItem.musicTrack</upnp:class><dc:creator>Item 1 Creator</dc:creator><upnp:album>Item 1 Album</upnp:album></item></DIDL-Lite>',
            'NumberReturned': '1',
            'TotalMatches': '10',
            'UpdateID': '1'}
        queue = moco.get_queue(start=8, max_items=32)
        moco.contentDirectory.Browse.assert_called_once_with([
            ('ObjectID', 'Q:0'),
            ('BrowseFlag', 'BrowseDirectChildren'),
            ('Filter', '*'),
            ('StartingIndex', 8),
            ('RequestedCount', 32),
            ('SortCriteria', '')
        ])
        assert queue is not None
        assert len(queue) == 1
        moco.contentDirectory.reset_mock()

    def test_soco_queue_size(self, moco):
        moco.contentDirectory.Browse.return_value = {
            'NumberReturned': '1',
            'Result': '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"><container id="Q:0" parentID="Q:" restricted="true" childCount="384"><dc:title>Queue Instance 0</dc:title><upnp:class>object.container.playlistContainer</upnp:class><res protocolInfo="x-rincon-queue:*:*:*">x-rincon-queue:RINCON_00012345678901400#0</res></container></DIDL-Lite>',
            'TotalMatches': '1',
            'UpdateID': '1'}
        queue_size = moco.queue_size
        moco.contentDirectory.Browse.assert_called_once_with([
            ('ObjectID', 'Q:0'),
            ('BrowseFlag', 'BrowseMetadata'),
            ('Filter', '*'),
            ('StartingIndex', 0),
            ('RequestedCount', 1),
            ('SortCriteria', '')
        ])
        assert queue_size == 384
        moco.contentDirectory.reset_mock()

    def test_join(self, moco_zgs):
        moco2 = mock.Mock()
        moco2.uid = "RINCON_000XXX1400"
        moco_zgs.join(moco2)
        moco_zgs.avTransport.SetAVTransportURI.assert_called_once_with(
            [('InstanceID', 0),
             ('CurrentURI', 'x-rincon:RINCON_000XXX1400'),
             ('CurrentURIMetaData', '')]
        )

    def test_unjoin(self, moco_zgs):
        moco_zgs.unjoin()
        moco_zgs.avTransport.BecomeCoordinatorOfStandaloneGroup\
            .assert_called_once_with([('InstanceID', 0)])

    def test_switch_to_line_in(self, moco_zgs):
        moco_zgs.avTransport.reset_mock()
        moco_zgs.switch_to_line_in()
        moco_zgs.avTransport.SetAVTransportURI.assert_called_once_with(
            [('InstanceID', 0),
                ('CurrentURI', 'x-rincon-stream:RINCON_000XXX1400'),
                ('CurrentURIMetaData', '')]
        )

        moco_zgs.avTransport.reset_mock()
        moco2 = mock.Mock()
        moco2.uid = "RINCON_000YYY1400"
        moco_zgs.switch_to_line_in(moco2)
        moco_zgs.avTransport.SetAVTransportURI.assert_called_once_with(
            [('InstanceID', 0),
             ('CurrentURI', 'x-rincon-stream:RINCON_000YYY1400'),
             ('CurrentURIMetaData', '')]
        )

    def test_switch_to_tv(self, moco_zgs):
        moco_zgs.avTransport.reset_mock()
        moco_zgs.switch_to_tv()
        moco_zgs.avTransport.SetAVTransportURI.assert_called_once_with(
            [('InstanceID', 0),
                ('CurrentURI', 'x-sonos-htastream:RINCON_000XXX1400:spdif'),
                ('CurrentURIMetaData', '')]
        )

    def test_is_playing_tv(self, moco):
        moco.avTransport.reset_mock()
        moco.avTransport.GetPositionInfo.return_value = {
            'TrackURI': 'x-sonos-htastream:RINCON_00012345678901234:spdif'
        }
        playing_tv = moco.is_playing_tv
        assert playing_tv
        moco.avTransport.GetPositionInfo.assert_called_once_with(
            [('InstanceID', 0),
             ('Channel', 'Master')]
        )

        moco.avTransport.GetPositionInfo.return_value = {
            'TrackURI': 'not-tv',
        }
        playing_tv = moco.is_playing_tv
        assert not playing_tv

    def test_is_playing_radio(self, moco):
        moco.avTransport.reset_mock()
        moco.avTransport.GetPositionInfo.return_value = {
            'TrackURI': 'x-rincon-mp3radio://example.com:80/myradio'
        }
        playing_tv = moco.is_playing_radio
        assert playing_tv
        moco.avTransport.GetPositionInfo.assert_called_once_with(
            [('InstanceID', 0),
             ('Channel', 'Master')]
        )

        moco.avTransport.GetPositionInfo.return_value = {
            'TrackURI': 'not-radio',
        }
        playing_tv = moco.is_playing_radio
        assert not playing_tv

    def test_is_playing_line_in(self, moco):
        moco.avTransport.reset_mock()
        moco.avTransport.GetPositionInfo.return_value = {
            'TrackURI': 'x-rincon-stream:blah-blah'
        }
        playing_tv = moco.is_playing_line_in
        assert playing_tv
        moco.avTransport.GetPositionInfo.assert_called_once_with(
            [('InstanceID', 0),
             ('Channel', 'Master')]
        )

        moco.avTransport.GetPositionInfo.return_value = {
            'TrackURI': 'not-line-in',
        }
        playing_tv = moco.is_playing_line_in
        assert not playing_tv

    def test_create_sonos_playlist(self, moco):
        playlist_name = "cool music"
        playlist_id = 1
        moco.avTransport.CreateSavedQueue.return_value = {
            'AssignedObjectID': 'SQ:{}'.format(playlist_id)
        }
        playlist = moco.create_sonos_playlist(playlist_name)
        moco.avTransport.CreateSavedQueue.assert_called_once_with(
            [('InstanceID', 0),
             ('Title', playlist_name),
             ('EnqueuedURI', ''),
             ('EnqueuedURIMetaData', '')]
        )
        assert playlist.title == playlist_name
        expected_uri = "file:///jffs/settings/savedqueues.rsq#{}".format(
            playlist_id)
        assert playlist.resources[0].uri == expected_uri
        assert playlist.parent_id == "SQ:"

    def test_create_sonos_playlist_from_queue(self, moco):
        playlist_name = "saved queue"
        playlist_id = 1
        moco.avTransport.SaveQueue.return_value = {
            'AssignedObjectID': 'SQ:{}'.format(playlist_id)
        }
        playlist = moco.create_sonos_playlist_from_queue(playlist_name)
        moco.avTransport.SaveQueue.assert_called_once_with(
            [('InstanceID', 0),
             ('Title', playlist_name),
             ('ObjectID', '')]
        )
        assert playlist.title == playlist_name
        expected_uri = "file:///jffs/settings/savedqueues.rsq#{}".format(
            playlist_id)
        assert playlist.resources[0].uri == expected_uri
        assert playlist.parent_id == "SQ:"

    def test_add_item_to_sonos_playlist(self, moco):
        moco.contentDirectory.reset_mock()

        playlist = mock.Mock()
        playlist.item_id = 7

        ressource = mock.Mock()
        ressource.uri = 'fake_uri'
        track = mock.Mock()
        track.resources = [ressource]
        track.uri = 'fake_uri'
        track.to_element.return_value = XML.Element('a')

        update_id = 100

        moco.contentDirectory.Browse.return_value = {
            'NumberReturned': '0',
            'Result': '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"></DIDL-Lite>',
            'TotalMatches': '0',
            'UpdateID': update_id
        }

        moco.add_item_to_sonos_playlist(track, playlist)
        moco.contentDirectory.Browse.assert_called_once_with([
            ('ObjectID', playlist.item_id),
            ('BrowseFlag', 'BrowseDirectChildren'),
            ('Filter', '*'),
            ('StartingIndex', 0),
            ('RequestedCount', 1),
            ('SortCriteria', '')
        ])
        moco.avTransport.AddURIToSavedQueue.assert_called_once_with(
            [('InstanceID', 0),
             ('UpdateID', update_id),
             ('ObjectID', playlist.item_id),
             ('EnqueuedURI', track.uri),
             ('EnqueuedURIMetaData', to_didl_string(track)),
             ('AddAtIndex', 4294967295)]
        )

    def test_soco_cross_fade(self, moco):
        moco.avTransport.GetCrossfadeMode.return_value = {
            'CrossfadeMode': '1'}
        assert moco.cross_fade
        moco.avTransport.GetCrossfadeMode.assert_called_once_with(
            [('InstanceID', 0)]
        )
        moco.cross_fade = False
        moco.avTransport.SetCrossfadeMode.assert_called_once_with(
            [('InstanceID', 0), ('CrossfadeMode', '0')]
        )

    def test_set_sleep_timer(self, moco):
        moco.avTransport.reset_mock()
        moco.avTransport.ConfigureSleepTimer.return_value = None
        result = moco.set_sleep_timer(None)
        assert result is None
        moco.avTransport.ConfigureSleepTimer.assert_called_once_with(
            [('InstanceID', 0),
             ('NewSleepTimerDuration', '')]
        )

        moco.avTransport.reset_mock()
        moco.avTransport.ConfigureSleepTimer.return_value = None
        result = moco.set_sleep_timer(7200)
        assert result is None
        moco.avTransport.ConfigureSleepTimer.assert_called_once_with(
            [('InstanceID', 0),
             ('NewSleepTimerDuration', '2:00:00')]
        )

        moco.avTransport.reset_mock()
        moco.avTransport.ConfigureSleepTimer.return_value = None
        result = moco.set_sleep_timer(0)
        assert result is None
        moco.avTransport.ConfigureSleepTimer.assert_called_once_with(
            [('InstanceID', 0),
             ('NewSleepTimerDuration', '0:00:00')]
        )

    @pytest.mark.parametrize('bad_sleep_time', ['BadTime', '00:43:23', '4200s',''])
    def test_set_sleep_timer_bad_sleep_time(self, moco, bad_sleep_time):
        with pytest.raises(ValueError):
            result = moco.set_sleep_timer(bad_sleep_time)

    def test_get_sleep_timer(self, moco):
        moco.avTransport.reset_mock()
        moco.avTransport.GetRemainingSleepTimerDuration.return_value = {
            'RemainingSleepTimerDuration': '02:00:00',
            'CurrentSleepTimerGeneration': '3',
        }
        result = moco.get_sleep_timer()
        assert result == 7200

        moco.avTransport.reset_mock()
        moco.avTransport.GetRemainingSleepTimerDuration.return_value = {
            'RemainingSleepTimerDuration': '',
            'CurrentSleepTimerGeneration': '0',
        }
        result = moco.get_sleep_timer()
        assert result == None


class TestContentDirectory:

    def test_remove_sonos_playlist_success(self, moco):
        moco.contentDirectory.reset_mock()
        moco.contentDirectory.return_value = True
        result = moco.remove_sonos_playlist('SQ:10')
        moco.contentDirectory.DestroyObject.assert_called_once_with(
            [('ObjectID', 'SQ:10')]
        )
        assert result


class TestRenderingControl:

    def test_soco_mute(self, moco):
        moco.renderingControl.GetMute.return_value = {'CurrentMute': '1'}
        assert moco.mute
        moco.renderingControl.reset_mock()
        moco.renderingControl.GetMute.return_value = {'CurrentMute': '0'}
        assert not moco.mute
        moco.renderingControl.GetMute.assert_called_once_with(
            [('InstanceID', 0), ('Channel', 'Master')]
        )
        moco.mute = False
        moco.renderingControl.SetMute.assert_called_once_with(
            [('InstanceID', 0), ('Channel', 'Master'),
                ('DesiredMute', '0')]
        )
        moco.renderingControl.reset_mock()
        moco.mute = True
        moco.renderingControl.SetMute.assert_called_once_with(
            [('InstanceID', 0), ('Channel', 'Master'),
                ('DesiredMute', '1')]
        )

    @pytest.mark.parametrize('volume', [1, 4, 10, 100])
    def test_soco_volume(self, moco, volume):
        moco.renderingControl.GetVolume.return_value = {
            'CurrentVolume': volume}
        assert moco.volume == volume
        moco.renderingControl.GetVolume.assert_called_once_with(
            [('InstanceID', 0), ('Channel', 'Master')]
        )
        moco.renderingControl.reset_mock()

    @pytest.mark.parametrize('vol_set, vol_called', [
        (-120, 0),
        (-10, 0),
        (0, 0),
        (5, 5),
        (99, 99),
        (100, 100),
        (110, 100),
        (300, 100)
    ])
    def soco_volume_set(self, moco, vol_set, vol_called):
        moco.volume = vol_set
        moco.renderingControl.SetVolume.assert_called_once_with(
            [('InstanceID', 0), ('Channel', 'Master'),
                ('DesiredVolume', vol_called)]
        )

    def test_soco_ramp_to_volume(self, moco):
        moco.renderingControl.RampToVolume.return_value = {'RampTime': '12'}
        ramp_time = moco.ramp_to_volume(15)
        moco.renderingControl.RampToVolume.assert_called_once_with(
            [('InstanceID', 0), ('Channel', 'Master'),
             ('RampType', 'SLEEP_TIMER_RAMP_TYPE'), ('DesiredVolume', 15),
             ('ResetVolumeAfter', False), ('ProgramURI', '')]
        )
        assert ramp_time == 12

    def test_soco_treble(self, moco):
        moco.renderingControl.GetTreble.return_value = {'CurrentTreble': '15'}
        assert moco.treble == 15
        moco.renderingControl.GetTreble.assert_called_once_with(
            [('InstanceID', 0), ('Channel', 'Master')]
        )
        moco.treble = '10'
        moco.renderingControl.SetTreble.assert_called_once_with(
            [('InstanceID', 0),
                ('DesiredTreble', 10)]
        )

    def test_soco_loudness(self, moco):
        moco.renderingControl.GetLoudness.return_value = {
            'CurrentLoudness': '1'}
        assert moco.loudness
        moco.renderingControl.GetLoudness.assert_called_once_with(
            [('InstanceID', 0), ('Channel', 'Master')]
        )
        moco.loudness = False
        moco.renderingControl.SetLoudness.assert_called_once_with(
            [('InstanceID', 0), ('Channel', 'Master'),
                ('DesiredLoudness', '0')]
        )


class TestDeviceProperties:

    def test_soco_status_light(self, moco):
        moco.deviceProperties.GetLEDState.return_value = {
            'CurrentLEDState': 'On'}
        assert moco.status_light
        moco.deviceProperties.GetLEDState.return_value = {
            'CurrentLEDState': 'Off'}
        assert not moco.status_light
        moco.deviceProperties.GetLEDState.assert_called_with()
        moco.status_light = False
        moco.deviceProperties.SetLEDState.assert_called_once_with(
            [('DesiredLEDState', 'Off')])
        moco.status_light = True
        moco.deviceProperties.SetLEDState.assert_called_with(
            [('DesiredLEDState', 'On')]
        )

    def test_soco_set_player_name(self, moco):
        moco.player_name = 'μИⅠℂ☺ΔЄ💋'
        moco.deviceProperties.SetZoneAttributes.assert_called_once_with(
            [('DesiredZoneName', 'μИⅠℂ☺ΔЄ💋'),
                ('DesiredIcon', ''),
                ('DesiredConfiguration', '')]
        )


class TestZoneGroupTopology:

    def test_soco_uid(self, moco_zgs):
        assert moco_zgs.uid == 'RINCON_000XXX1400'

    def test_soco_is_visible(self, moco_zgs):
        assert moco_zgs.is_visible

    def test_soco_is_bridge(self, moco_zgs):
        assert not moco_zgs.is_bridge

    def test_soco_is_coordinator(self, moco_zgs):
        assert moco_zgs.is_coordinator

    def test_all_groups(self, moco_zgs):
        groups = moco_zgs.all_groups
        assert len(groups) == 3
        # Check 3 unique groups
        assert len(set(groups)) == 3
        for group in groups:
            assert isinstance(group, ZoneGroup)

    def test_all_groups_have_coordinator(self, moco_zgs):
        for group in moco_zgs.all_groups:
            print("GROUP", group.coordinator)
            print(group)
            assert group.coordinator is not None

    def test_group(self, moco_zgs):
        assert isinstance(moco_zgs.group, ZoneGroup)
        assert moco_zgs in moco_zgs.group

    def test_all_zones(selfself, moco_zgs):
        zones = moco_zgs.all_zones
        assert len(zones) == 4
        assert len(set(zones)) == 4
        for zone in zones:
            assert isinstance(zone, SoCo)
        assert moco_zgs in zones

    def test_visible_zones(selfself, moco_zgs):
        zones = moco_zgs.visible_zones
        assert len(zones) == 4
        assert len(set(zones)) == 4
        for zone in zones:
            assert isinstance(zone, SoCo)
        assert moco_zgs in zones

    def test_group_label(selfself, moco_zgs):
        g = moco_zgs.group
        # Have to mock out group members zone group state here since
        # g.members is parsed from the XML.
        for speaker in g.members:
            speaker.zoneGroupTopology.GetZoneGroupState.return_value = {
                'ZoneGroupState': ZGS
            }
        assert g.label == "Kitchen, Living Room"

    def test_group_short_label(selfself, moco_zgs):
        g = moco_zgs.group
        # Have to mock out group members zone group state here since
        # g.members is parsed from the XML.
        for speaker in g.members:
            speaker.zoneGroupTopology.GetZoneGroupState.return_value = {
                'ZoneGroupState': ZGS
            }
        assert g.short_label == "Kitchen + 1"


def test_only_on_master_true(moco_only_on_master):
    with mock.patch('pysonos.SoCo.is_coordinator', new_callable=mock.PropertyMock) as is_coord:
        is_coord.return_value = True
        moco_only_on_master.play()
        is_coord.assert_called_once_with()


def test_not_on_master_false(moco_only_on_master):
    with mock.patch('pysonos.SoCo.is_coordinator', new_callable=mock.PropertyMock) as is_coord:
        is_coord.return_value = False
        with pytest.raises(SoCoSlaveException):
            moco_only_on_master.play()
        is_coord.assert_called_once_with()
